with Ada.Text_IO, Flts;
use Ada.Text_IO, Flts;

generic

-------------------- begin_include reps Rep

  type Rep is private;  --- used to define intervals, balls, ...

  with function IsZero(R: Rep) return Boolean is <>;
  with procedure SetZero(R: in out Rep) is <>;
  with function Sign(R: Rep) return Integer is <>;
  with function "abs"(R: Rep) return Rep is <>;
  with function Min(R1,R2: Rep) return Rep is <>;
  with function Max(R1,R2: Rep) return Rep is <>;
  with procedure Swap(R1,R2: in out Rep) is <>;

  with function Compare(R1,R2: Rep) return Integer is <>;
  with function "<"(R1,R2: Rep) return Boolean is <>;
  with function "<="(R1,R2: Rep) return Boolean is <>;
  with function "="(R1,R2: Rep) return Boolean is <>;
  with function ">="(R1,R2: Rep) return Boolean is <>;
  with function ">"(R1,R2: Rep) return Boolean is <>;

  with function "-"(R: Rep) return Rep is <>;
  with function "+"(R1,R2: Rep) return Rep is <>;
  with function "-"(R1,R2: Rep) return Rep is <>;
  with function "*"(R1,R2: Rep) return Rep is <>;
  with function "/"(R1,R2: Rep) return Rep is <>;
  with function "**"(R: Rep; N: Integer) return Rep is <>;
  with function "*"(F: Flt; R: Rep) return Rep is <>;
  with function Sqrt(R: Rep) return Rep is <>;

  with function Upper(R: Rep) return Flt is <>;
  with function Lower(R: Rep) return Flt is <>;
  with function Approx(R: Rep) return Flt is <>;
  with function ApproxPut(R: Rep) return FltVector is <>;
  with function ApproxGet(V: FltVector) return Rep is <>;
  with function Val(I: Integer) return Rep is <>;
  with function Val(F: Flt) return Rep is <>;
  with function ValDec(S: String) return Rep is <>;
  with function ValHex(S: String) return Rep is <>;
  with function DecStr(R: Rep; Digs: Natural := 0) return String is <>; -- default, if NDig=0
  with function HexStr(R: Rep; Digs: Natural := 0) return String is <>; -- full, if NDig=0
  with function Rounded_Last_Conversion return Boolean is <>;

  with procedure Set_Precision(Dummy: in Rep; Binary_Digs: in Positive) is <>;
  with function Get_Precision(R: Rep) return Positive is <>;
  with procedure Set_Rounding_Up(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Down(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Nearest(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Truncate(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Relaxed(Dummy: in Rep) is <>;

--- approximate
  with function Num_Exp(R: Rep) return Rep is <>;
  with function Num_Log(R: Rep) return Rep is <>;
  with function Num_Cos(R: Rep) return Rep is <>;
  with function Num_Sin(R: Rep) return Rep is <>;

-------------------- end_include reps Rep

package Balls is  -- uses permanent rounding_up mode

  pragma Elaborate_Body;

  type Ball is private;

--- basic
  function "="(S1,S2: Ball) return Boolean;
  function IsNumeric(S: Ball) return Boolean;
  function IsZero(S: Ball) return Boolean;
  procedure SetZero(S: in out Ball);
  function Copy(S: Ball) return Ball;
  procedure Enclosure(S: in out Ball; Binary_Digs: in Positive);
  function Approx(S: Ball) return Flt;
  function FPut(S: Ball) return FltVector;
  function FGet(V: FltVector) return Ball;
  function Scal(K: Integer) return Ball;
  function Scal(R: Flt) return Ball;
  function Scal(R1,R2: Flt) return Ball;
  procedure Swap(S1,S2: in out Ball);

--- intervals
  function Sup(S: Ball) return Flt;
  function Inf(S: Ball) return Flt;
  function Disjoint(S1,S2: Ball) return Boolean;
  function "<"(S1,S2: Ball) return Boolean;
  function "<="(S1,S2: Ball) return Boolean;
  function ">"(S1,S2: Ball) return Boolean;
  function ">="(S1,S2: Ball) return Boolean;
  function Min(S1,S2: Ball) return Ball;
  function Max(S1,S2: Ball) return Ball;

--- sets
  function Ball0(S: Ball) return Ball;
  function Ball0(R: Flt) return Ball;
  function Center(S: Ball) return Ball;
  procedure ResetCenter(S: in out Ball);
  function Subset(S1,S2: Ball) return Boolean;
  function Union(S1,S2: Ball) return Ball;
  function Intersection(S1,S2: Ball) return Ball;

--- ops
  function Norm(S: Ball) return Ball;
  function "abs"(S: Ball) return Ball renames Norm;
  function MaxNorm(S: Ball) return Flt;
  procedure Neg(S: in out Ball);
  function "-"(S: Ball) return Ball;
  procedure Add(S: in Ball; T: in out Ball);
  function "+"(S,T: Ball) return Ball;
  procedure Sub(S: in Ball; T: in out Ball);
  function "-"(S,T: Ball) return Ball;
  procedure Mult(R: in Flt; S: in out Ball);
  function "*"(R: Flt; S: Ball) return Ball;
  function Sqr(S: Ball) return Ball;
  procedure Mult(S: in Ball; T: in out Ball);
  function "*"(S,T: Ball) return Ball;
  procedure AddProd(S1,S2: in Ball; S3: in out Ball);
  function Inv(S: Ball) return Ball;
  function "/"(S,T: Ball) return Ball;
  function "**"(S: Ball; I: Integer) return Ball;

--- fun
  function Sqrt(S: Ball) return Ball;
  function Exp(S: Ball) return Ball;
  function Log(S: Ball) return Ball;
  function Cos(S: Ball) return Ball;
  function Sin(S: Ball) return Ball;
  function Pi return Ball;

--- io
  procedure ShowType(S: in Ball);
  procedure Show1(N: in String; S: in Ball; NewLine: in Boolean := True);
  procedure Show2(N: in String; S1,S2: in Ball; NewLine: in Boolean := True);
  procedure Put(F: in File_Type; S: in Ball; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; S: in out Ball; Decimal: in Boolean := False);
  procedure Write(Name: in String; S: in Ball; Decimal: in Boolean := False);
  procedure Read(Name: in String; S: in out Ball; Decimal: in Boolean := False);

--- misc
  procedure Set_Active_Mode(Dummy: in Ball);
  procedure Set_Wait_Mode(Dummy: in Ball);
  procedure Set_New_Precision(Dummy: in Ball; Binary_Digs: in Positive);
  function Get_Precision(S: Ball) return Positive;

private

  type Ball is
    record
      C: Rep;     -- center of ball
      R: Rep;    -- radius of ball
    end record;

  function RSup(S: Ball) return Rep;
  function RInf(S: Ball) return Rep;
  function Scal(R1,R2: Rep) return Ball;
  function ExpUp(R: Rep) return Rep;
  function ExpDown(R: Rep) return Rep;
  procedure CosSin(S: in Ball; Sc,Ss: in out Ball);
  function Enlarge(R: in Rep) return Ball;

  Undefined,Not_Allowed: exception;
  Active: Boolean := True;
  Precision: Positive := Get_Precision(Val(0));

end Balls;
