with Ada.Text_IO, Ints, Flts, Messages, Vectors, Vectors.Ops;
use Ada.Text_IO, Ints, Flts, Messages;

generic

  type Scalar is private;

  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with function Ball0(R: Flt) return Scalar is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Sup(S: Scalar) return Flt is <>;
  with function Intersection(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function "="(S,T: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with function MaxNorm(S: Scalar) return Flt is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  with package SV is new Vectors (Scalar => Scalar);
  with package SV_Ops is new SV.Ops;
  use SV, SV_Ops;

  type Fun is private;

  with procedure SetZero(Template: in Fun; F: in out Fun) is <>;
  with function Norm(F: Fun) return Scalar is <>;
  with procedure Pack(F: in Fun; V: in out Vector) is <>;
  with procedure UnPack(V: in Vector; Template: in Fun; F: in out Fun) is <>;
  with procedure Add(F: in Fun; G: in out Fun) is <>;
  with procedure Sum(F,G: in Fun; H: in out Fun) is <>;
  with procedure Sub(F: in Fun; G: in out Fun) is <>;
  with procedure Diff(F,G: in Fun; H: in out Fun) is <>;
  with function Deg(K: Positive) return Natural is <>;
  with procedure AddMode(K: in Positive; S: in Scalar; F: in out Fun) is <>;
  with procedure AddHigher(K: in Positive; S: in Scalar; F: in out Fun) is <>;
  with procedure AddBall(S: in Scalar; F: in out Fun) is <>;
  with procedure Write(Name: in String; H: in Fun; Decimal: in Boolean := False) is <>;

package Contraction is

  generic
    with procedure Op(F: in Fun; G: in out Fun);
  function HighNorm(Template: Fun; K0: Positive; Target: Radius) return Radius;

  generic
    with function HighDeg(K: Positive) return Boolean;
    with procedure LowOp(F: in Fun; G: in out Fun);
    with procedure HighOp(F: in Fun; G: in out Fun);
  function OperatorNorm(Template: Fun; Small,Target: Radius) return Scalar;

--- prodecures Num* are meant to be used with numeric scalars

  generic
    with procedure Op(F: in Fun; G: in out Fun);
  procedure NumOpMatrix(Template: Fun; A: in out Matrix);

  function NewDeg(K: Positive) return Boolean;
  procedure DiagProd(W: in Vector; F: in Fun; G: in out Fun);
  procedure Prod(A: in Matrix; F: in Fun; G: in out Fun);
  function OpNorm(Template: Fun; M: Matrix) return Scalar;

  subtype Param is Vector;
  type Map is access procedure(P: in Param; F: in Fun; G: in out Fun);
  type DMap is access procedure(P: in Param; F: in Fun; H: in out Fun);

  procedure NumDerMatrix(DY: in DMap; P: in Param; F: in Fun; A: in out Matrix);
  procedure NumContrMatrix(DY: in DMap; P: in Param; F: in Fun; A: in out Matrix);
  procedure Contract(Y: in Map; P: in Param; F: in Fun; A: in Matrix; H: in out Fun; Err: out Flt);
  procedure Contract0(Y: in Map; P: in Param; A: in Matrix; F: in out Fun; Err: out Flt);
  procedure NumFix(Y: in Map; DY: in DMap; P: in Param; F: in out Fun; A: in out Matrix; Eps: in Radius);
  function DContractNorm(DY: DMap; P: Param; F: Fun; A: Matrix; Target: Radius) return Scalar;
  procedure CheckFix(Y: in Map; DY: in DMap; P: in Param; A: in Matrix; F: in out Fun; KMax: in Radius);

end Contraction;
