with Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO, Symmetries, Numerics, Square;
use Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO;

--with Long_Floats; use Long_Floats;
with MPFR.Floats; use MPFR.Floats;

procedure ContrLoop is

--  type NRep is new Rep;  -- rep used for numeric computations
--  type SRep is new Rep;  -- rep used to define Scalar

  type NRep is new MPFloat;  -- rep used for numeric computations
  type SRep is new MPFloat;  -- rep used to define Scalar

  package Scalars is new Numerics (Rep => SRep);
  use Scalars;
  type Scalar is new Numeric;

  Deg:   constant Positive := GetArg;
  IName: constant String   := GetArg;
  OName: constant String   := GetArg;

  package SQ is new Square (Deg => Deg, NRep => NRep, Scalar => Scalar);
  use SQ;
  use SConstants, SVectors, SFouriers2, SContr;
  subtype Fun is SFourier2;

  DC:     constant Integer := 9;           -- used in contraction
  LSize:  constant Integer := 150;         -- used for eigenvalues/vectors
  Eps:    constant Radius  := 1.0E-20;     -- for approx solutions
  Rad:    constant Radius  := 1.0/8.0;     -- domain
  Lambda: constant Flt     := 145.0/256.0;
  P:      constant SParam2 := (Scal(Lambda),ScalZero);

  procedure ShowParam is
  begin
    Show_Param;
    Show1("    DC =",DC);
    Show1(" LSize =",LSize);
    Show1("   Eps =",Eps);
    Show1("Lambda =",Lambda);
    Show1("   Rad =",Rad);
  end ShowParam;

  F0,F1: Fun;

begin
  Verbosity := 1;
  Set_Precision(128);
  ShowParam;
  Symmetries.Reflect := Symmetries.Trivial'Access;     -- no symmetry

  Read(IName,F0);
  F0.T.R := (Rad,Rad);
  loop
    F1 := ContrFixA(P(1),F0,DC,Eps);
    Write(OName,F1);
    Show1("norm",Norm(F1));
    MapA(P,F1,F0);
    Show1(" err",Norm(F1-F0));
    F0 := F1;
  end loop;

end ContrLoop;
