with Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO;
use Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO;
with NumFouriers2;

--with Long_Floats; use Long_Floats;
with MPFR.Floats; use MPFR.Floats;

procedure DMapMatrix is

--  type NRep is new Rep;
  type NRep is new MPFloat;

  Deg: constant Positive := 100;

  package NF is new NumFouriers2 (Deg1 => Deg, Deg2 => Deg, NRep => NRep);
  use NF;
  use Numbers, NConstants, NVectors, NVectors_Ops, NVectors_Num, NFouriers2;
  use Contr;

  LSize:  constant Integer := 400;         -- used for eigenvalues/vectors
  Eps:    constant Radius  := 1.0E-15;
  Rad:    constant Radius  := Half**60;
  Lambda: constant Flt     := Flt(145)/Flt(256);
  P:      constant NParam(1..2)  := (Scal(Lambda),ScalZero);

  procedure ShowParam is
  begin
    ShowType(ScalZero);
    Show1("N-Bits =",Get_Precision(ScalZero));
    Show1("   Deg =",Deg);
    Show1(" LSize =",LSize);
    Show1("   Eps =",Eps);
    Show1("Lambda =",Lambda);
    Show1("   Rad =",Rad);
  end ShowParam;


  F,G: NFourier2;
  W,VJ: NVector(1..LSize);
  A,V: NMatrix(1..LSize,1..LSize);
  T: NSimDiag(LSize);

begin
  Verbosity := 2;
  Set_Precision(ScalZero,128);
  ShowParam;

  Read("approxfix-200.data",F);

  NumDerMatrix(DMapA'Access,P,F,A);
  Sobolize(A);
  Symmetrize(A);
  Write("dmapa.data",A);

--return;
--  Read("dmapa.data",A);

  Eigen(A,W,V);
  Show0("Eigenvalues");
  Show1("W ",W);

  Show("ortho err ",Sup(Norm1(OrthoErr(V))));
  Orthogonalize(V);
  for J in 1 .. 3 loop
    New_Line;
    Show0("Eigenvector" & Strng(J,3));
    VJ := GetColumn(J,V);
    UnSobolize(VJ);
    Show1("V ",VJ);
    UnPack(VJ,F,G);
    Write("evec" & Strng(J) & ".data",G);
  end loop;

  V := (V*Diag(W))*Transpose(V);
  Sub(V,A);
  Show("quality ",Sup(Norm1(A)));

end DMapMatrix;
