with Ints, Ints.IO;
use Ints, Ints.IO;

pragma Optimize (Off);

package body FltStr is

  package Flt_IO is new Ada.Text_IO.Float_IO(Flt);

  function ValHex(S: String) return Flt is
    NHex: constant Integer := 1+Flt'Machine_Mantissa/4;
    N,D,D1,DN,E1,E2,Ed1,Es2,Ms: Integer := 0;
    R: Flt;
  begin
    R := Flt(0);
    for K in S'Range loop
      if S(K)=' ' then
        null;
      elsif S(K)='+' then
        if Ms=0 then Ms := 1; else Es2 := 1; end if;
      elsif S(K)='-' then
        if Ms=0 then Ms := -1; else Es2 := -1; end if;
      elsif Es2=0 then
        if Ms=0 then Ms := 1; end if;
        if S(K)='.' then
          Ed1 := -1;
        else
          N := N+1;
          D := Val(S(K),15);
          if N <= NHex then
            if N=1 then D1 := D; elsif N=NHex then DN := D; end if;
            R := Flt(16)*R+Flt(D);
            E1 := E1+Ed1;
          elsif D>0 then
            DN := 31;
          end if;
        end if;
      else
        E2 := 16*E2+Val(S(K),15);
      end if;
    end loop;
    R := Flt(Ms)*Flt'Scaling(R,4*(E1+Es2*E2));
    Rounded_Last := (Strip0(16*D1+DN,2) >= 16);
    return R;
  end ValHex;

  function FltCompose(R: in Flt; E: in Integer) return Flt is
--- workaround for a bug in Aonix ObjectAda 8.2
  begin
    return Flt'Compose(Flt'Fraction(R),E);
  end FltCompose;

  function HexStr(R: Flt; Digs: Natural := 0) return String is
    NHex: constant Integer := 1+Flt'Machine_Mantissa/4;
    NMax: Integer;
  begin
    Rounded_Last := False;
    if R=0.0 then return " 0.0+0"; end if;
    if Digs >= 2 then
      NMax := IMin(Digs,NHex)-1;
    elsif Digs=0 then
      NMax := NHex-1;
    else
      NMax := 1;
    end if;
    declare
      E:    constant Integer := Flt'Exponent(R)-1;
      F:    constant Integer := E mod 4;
      M:    constant Flt     := FltCompose(Abs(R),F+1);
      D: Integer;
      L: Flt;
      S: String(1 .. NMax+3);
    begin
      if R<0.0 then S(1) := '-'; else S(1) := '+'; end if;
      L := Flt'Leading_Part(M,F+1);
      S(2) := UpDig(Integer(Flt'Floor(L)));
      S(3) := '.';
      for N in 1 .. NHex-1 loop
        L := Flt'Scaling(M-Flt'Leading_Part(M,F-3+4*N),4*N);
        D := Integer(Flt'Floor(L));
        if N <= NMax then
          S(N+3) := UpDig(D);
        elsif D>0 then
          Rounded_Last := True;
          exit;
        end if;
      end loop;
      return S & Image((E-F)/4,16,True);
    end;
  end HexStr;

  function DecStr(R: Flt; Digs: Natural := 0) return String is
    Fore: constant Field := 2; -- default
    Exp:  constant Field := 4; -- default
    Aft: Field := Flt'Digits;  -- default
  begin
    if Digs>1 then Aft := Digs-1;
    elsif Digs=1 then Aft := 1; end if;
    declare
      S: String(1 .. Fore+1+Aft+1+Exp);
    begin
      Flt_IO.Put(S,R,Aft,Exp);
      Rounded_Last := True;
      return S;
    end;
  end DecStr;

end FltStr;
