with Ada.Text_IO, Ints, Flts, Symmetries;
use Ada.Text_IO, Ints, Flts, Symmetries;
with Constants, Vectors, Vectors.Ops, Vectors.Num, Contraction;

pragma Elaborate_All (Constants);

generic

  Deg1,Deg2: in Positive;

-------------------- begin_include scalars Scalar

  type Scalar is private;

--- basic
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with function Copy(S: Scalar) return Scalar is <>;
  with procedure Enclosure(S: in out Scalar; Binary_Digs: in Positive) is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with function FPut(R: Scalar) return FltVector is <>;
  with function FGet(V: FltVector) return Scalar is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function Scal(R1,R2: Flt) return Scalar is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

--- intervals
  with function Sup(S: Scalar) return Flt is <>;
  with function Inf(S: Scalar) return Flt is <>;
  with function Disjoint(S1,S2: Scalar) return Boolean is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;

--- sets
  with function Ball0(S: Scalar) return Scalar is <>;
  with function Ball0(R: Flt) return Scalar is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ResetCenter(S: in out Scalar) is <>;
  with function Subset(S1,S2: Scalar) return Boolean is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with function Intersection(S1,S2: Scalar) return Scalar is <>;

--- ops
  with function Norm(S: Scalar) return Scalar is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Flt is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(S: in Scalar; T: in out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with procedure Sub(S: in Scalar; T: in out Scalar) is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Mult(S: in Scalar; T: in out Scalar) is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

--- fun
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Pi return Scalar is <>;

--- io
  with procedure ShowType(S: in Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;
  with procedure Write(Name: in String; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Read(Name: in String; S: in out Scalar; Decimal: in Boolean := False) is <>;

--- misc
  with procedure Set_Active_Mode(Dummy: in Scalar) is <>;
  with procedure Set_Wait_Mode(Dummy: in Scalar) is <>;
  with procedure Set_New_Precision(Dummy: in Scalar; Binary_Digs: in Positive) is <>;
  with function Get_Precision(S: Scalar) return Positive is <>;
-------------------- end_include scalars Scalar

  with package SC is new Constants (Scalar => Scalar);
  with package SV is new Vectors (Scalar => Scalar);
  with package SV_Ops is new SV.Ops;
  with package SV_Num is new SV.Num;
  use SC, SV, SV_Ops;

package Fouriers2 is

  pragma Elaborate_Body;

  DegE1: constant Positive := 2*Deg1;
  DegE2: constant Positive := 2*Deg2;
  DegE:  constant Positive := IMax(DegE1,DegE2);

  subtype Freq1  is Integer range 0 .. Deg1;
  subtype Freq2  is Integer range 0 .. Deg2;
  subtype FreqE1 is Integer range 0 .. DegE1;
  subtype FreqE2 is Integer range 0 .. DegE2;

  subtype ScalVec is SV.Vector;
  subtype ScalDiag is SV.Diag;
  subtype ScalMat is SV.Matrix;
  subtype ScalSimDiag is SV_Ops.SimDiag;
  subtype ScalLU is SV_Ops.LUData;
  subtype FCoeff is ScalMat(Freq1,Freq2);
  subtype FErr is RadMatrix(FreqE1,FreqE2);

  type FType is
    record
      P: ParPair;  --- even (0) or odd (1)
      R: RadPair;  --- domain parameter
    end record;

  type Fourier2 is
    record
      T: FType;
      C: FCoeff;      --- coefficients
      E: FErr;        --- higher order error bounds
    end record;

--- basic
  function IsZero(F: Fourier2) return Boolean;
  function IsPoly(F: Fourier2) return Boolean;
  procedure SetZero(T: in FType; F: in out Fourier2);
  procedure SetZero(Template: in Fourier2; F: in out Fourier2);
  procedure SetZero(P: in ParPair; R: in RadPair; F: in out Fourier2);
  function Norm(R: RadPair; C: FCoeff) return Scalar;
  function Norm(E: FErr) return Scalar;
  function Norm(R: RadPair; F: Fourier2) return Scalar;
  function Norm(F: Fourier2) return Scalar;
  function IProd(F,G: in Fourier2) return Scalar;
  function Val(X,Y: Scalar; F: Fourier2) return Scalar;
  function Coeff(M,N: Natural; F: Fourier2) return Scalar;
  procedure AddCoeff(M,N: in Natural; S: in Scalar; F: in out Fourier2);
  procedure AddBall(M,N: in Natural; S: in Scalar; F: in out Fourier2);
  procedure AddBall(S: in Scalar; F: in out Fourier2);

--- ops
  procedure Neg(F: in out Fourier2);
  function "-"(F: Fourier2) return Fourier2;
  procedure Add(F: in Fourier2; G: in out Fourier2);
  procedure Sum(F,G: in Fourier2; H: in out Fourier2);
  function "+"(F,G: Fourier2) return Fourier2;
  procedure Sub(F: in Fourier2; G: in out Fourier2);
  procedure Diff(F,G: in Fourier2; H: in out Fourier2);
  function "-"(F,G: Fourier2) return Fourier2;
  procedure Mult(S: in Scalar; F: in out Fourier2);
  procedure Prod(S: in Scalar; F: in Fourier2; G: in out Fourier2);
  function "*"(S: Scalar; F: Fourier2) return Fourier2;
  procedure AddProd(S: in Scalar; F: in Fourier2; G: in out Fourier2);
  procedure Mult(F: in Fourier2; G: in out Fourier2);
  procedure Prod(F,G: in Fourier2; H: in out Fourier2);
  function "*"(F,G: Fourier2) return Fourier2;
  procedure AddProd(F1,F2: in Fourier2; F3: in out Fourier2);

--- io
  procedure Show1(S: in String; F: in Fourier2);
  procedure Show2(S: in String; F1,F2: in Fourier2);
  procedure Put(F: in File_Type; H: in Fourier2; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; H: in out Fourier2; Decimal: in Boolean := False);
  procedure Write(Name: in String; H: in Fourier2; Decimal: in Boolean := False);
  procedure Read(Name: in String; H: in out Fourier2; Decimal: in Boolean := False);
  procedure NumRead(Name: in String; H: in out Fourier2; Decimal: in Boolean := False);
  procedure Plot(F: in Fourier2; L: Positive := 50);
  procedure GuessRho(F: Fourier2);

--- symmetric modes
  procedure Reflect(F: in out Fourier2);
  procedure Symmetrize(F: in out Fourier2);
  function ModeCoeff(K: Positive; C: FCoeff) return Scalar;
  procedure ModeSplit(K: in Positive; C: in out FCoeff; S: in out Scalar);
  function ModeCoeff(K: Positive; F: Fourier2) return Scalar;
  procedure AddMode(K: in Positive; S: in Scalar; C: in out FCoeff);
  procedure AddMode(K: in Positive; S: in Scalar; F: in out Fourier2);
  procedure AddHigher(K: in Positive; R: in Flt; E: in out FErr);
  procedure AddHigher(K: in Positive; S: in Scalar; F: in out Fourier2);

  function SomeDeg(C: FCoeff) return Positive;
  procedure Pack(C: in FCoeff; V: in out ScalVec);
  procedure PackSplit(C: in out FCoeff; V: in out ScalVec);
  procedure Pack(F: in Fourier2; V: in out ScalVec);
  procedure PackSplit(F: in out Fourier2; V: in out ScalVec);
  procedure UnPack(V: in ScalVec; C: in out FCoeff);
  procedure UnPack(V: in ScalVec; Template: in Fourier2; F: in out Fourier2);
  function UnPack(V: ScalVec; Template: Fourier2) return Fourier2;

--- examples
  function AntiBumpsX(R: RadPair) return Fourier2;
  function AntiBumpsY(R: RadPair) return Fourier2;
  function AntiBumpsDiag(R: RadPair) return Fourier2;
  function MidPeak(R: RadPair; N: Positive) return Fourier2;    -- peak at center
  function MidHole(R: RadPair; N: Natural) return Fourier2;     -- hole at center
  function EdgePeaks(R: RadPair; N: Positive) return Fourier2;  -- peaks on edge centers
  function CornerPeaks(R: RadPair; N: Natural) return Fourier2; -- peaks in corners
  function MValley(R: RadPair; N: Positive) return Fourier2;    -- zero on main diag
  function PValley(R: RadPair; N: Positive) return Fourier2;    -- zero on other diag
  function HVValley(R: RadPair; N: Natural) return Fourier2;    -- zero horiz and vert
  function MRidge(R: RadPair; N: Natural) return Fourier2;      -- ridge on main diag
  function PRidge(R: RadPair; N: Natural) return Fourier2;      -- ridge on other diag
  function LR_Peaks(R: RadPair; N: Positive) return Fourier2;   -- peaks left & right
  function BT_Peaks(R: RadPair; N: Positive) return Fourier2;   -- peaks bottom & top
  function MPeaks(R: RadPair; N: Positive) return Fourier2;     -- peaks on main diag
  function PPeaks(R: RadPair; N: Positive) return Fourier2;     -- peaks on other diag
  function PMPeaks(R: RadPair; N: Positive) return Fourier2;    -- peaks on both diag
  function LR_Holes(R: RadPair; N: Natural) return Fourier2;    -- holes left & right
  function BT_Holes(R: RadPair; N: Natural) return Fourier2;    -- holes bottom & top
  function MHoles(R: RadPair; N: Natural) return Fourier2;      -- holes on main diag
  function PHoles(R: RadPair; N: Natural) return Fourier2;      -- holes on other diag
  function PMHoles(R: RadPair; N: Positive) return Fourier2;    -- holes on both diag
  function HVRidge(R: RadPair; N: Natural) return Fourier2;     -- ridge horiz and vert
  function EdgeRidge(R: RadPair; N: Natural) return Fourier2;   -- ridge at boundary
  function Peaks8(R: RadPair; N: Positive) return Fourier2;     -- 8 peaks

--- maps
  function Perturb1(R: RadPair; L: Scalar) return Fourier2;
  function Perturb2(R: RadPair; L: Scalar) return Fourier2;
  function Perturb3(R: RadPair; L: Scalar) return Fourier2;
  function Perturb4(R: RadPair; L: Scalar) return Fourier2;
  function Perturb5(R: RadPair; L: Scalar) return Fourier2;
  function Perturb6(R: RadPair; L: Scalar) return Fourier2;
  function Perturb7(R: RadPair; L: Scalar) return Fourier2;
  function Perturb8(R: RadPair; L: Scalar) return Fourier2;
  function Perturb9(R: RadPair; L: Scalar) return Fourier2;
  function Perturb10(R: RadPair; L: Scalar) return Fourier2;
  function Perturb11(R: RadPair; L: Scalar) return Fourier2;
  function Perturb12(R: RadPair; L: Scalar) return Fourier2;
  function Perturb13(R: RadPair; L: Scalar) return Fourier2;
  function YPerturb1(R: RadPair; L: Scalar) return Fourier2;
  function Perturb(R: RadPair; L: Scalar) return Fourier2 renames Perturb13;

  function LowSqrtNegLap(Last: Positive) return ScalDiag;
  procedure Sobolize(V: in out ScalVec);
  procedure Sobolize(A: in out ScalMat);
  procedure UnSobolize(V: in out ScalVec);
  procedure UnSobolize(A: in out ScalMat);
  procedure InvNegLaplace(F: in out Fourier2);
  procedure MapA(P: in ScalVec; F: in Fourier2; G: in out Fourier2);
  procedure DMapA(P: in ScalVec; F: in Fourier2; H: in out Fourier2);

  package Contr is new Contraction (Scalar => Scalar, SV => SV, SV_Ops => SV_Ops, Fun => Fourier2);
  subtype Param is Contr.Param;
  subtype Map is Contr.Map;
  subtype DMap is Contr.DMap;

--- numerical
  procedure NumEigenData(DY: in DMap; P: in Param; F: in Fourier2; T: in out ScalSimDiag);

--- bifurcation related maps
  function Compo1(F,F0: Fourier2) return Scalar;
  function Proj1(F,F0: Fourier2) return Fourier2;
  procedure BifMap(Y: in Map; V: ScalVec; F: in Fourier2; G: in out Fourier2);
  procedure DBifMap(DY: in DMap; V: ScalVec; F: in Fourier2; H: in out Fourier2);
  procedure BifMapA(V: ScalVec; F: in Fourier2; G: in out Fourier2);
  procedure DBifMapA(V: ScalVec; F: in Fourier2; H: in out Fourier2);

  Zero_FCoeff: constant FCoeff    := ZeroMatrix(0,Deg1,0,Deg2);
  Zero_FErr:   constant FErr      := ZeroRadMatrix(0,DegE1,0,DegE2);

private

  procedure Add(F: in FErr; G: in out FErr);
  procedure Mult(S: in Scalar; E: in out FErr);
  procedure AddProd(S: in Scalar; F: in FErr; G: in out FErr);
  procedure FillZero(F: in out Fourier2);
  function NotExp(X: Flt) return Scalar;
  procedure Powers(S: in Scalar; W: in out ScalVec);
  procedure UpdateNormFacs(R: in RadPair);
  function IProd(C1,C2: FCoeff) return Scalar;
  function CosSin(P: Parity; S: Scalar) return Scalar;
  function Coeff(M,N: Natural; R: RadPair; E: FErr) return Scalar;
  procedure AddBall(M,N: in Natural; R: in Flt; E: in out FErr);
  procedure CheckType(T1: in FType; T2: in out FType; SameParity: out Boolean);
  procedure Signs(P1,P2: in Parity; Sum,DiffPlus,DiffMinus: out Flt);
  procedure AddProdCC(F1,F2: in Fourier2; F3: in out Fourier2);
  procedure AddProdCE(F1,F2: in Fourier2; F3: in out Fourier2);
  procedure AddProdEE(F1,F2: in Fourier2; F3: in out Fourier2);
  function "*"(P1,P2: ParPair) return ParPair;
  procedure Normalize(F: in out Fourier2);
  procedure Show1(S: in String; T: in FType);
  function EffDeg1(C: FCoeff) return Integer;
  function EffDeg2(C: FCoeff) return Integer;
  procedure EffDegs0(C: in FCoeff; D1,D2: out Integer);
  function EffDeg1(E: FErr) return Integer;
  function EffDeg2(E: FErr) return Integer;
  procedure EffDegs0(E: in FErr; D1,D2: out Integer);
  procedure CReflect(C: in out FCoeff);
  procedure EReflect(E: in out FErr);
  procedure CSymm(C: in out FCoeff);
  procedure ESymm(E: in out FErr);
  procedure DistErr(D: in Natural; F: in out Fourier2);
  procedure DMapA(H: in out Fourier2);

  LastPair: RadPair := (One,One);
  NormFacs1,NormFacs2: ScalVec(-DegE .. DegE);
  NormFacs: ScalMat(FreqE1,FreqE2);

  Saved: Boolean := False;
  LSave: Scalar;
  FSave,GSave: Fourier2;

  Nop: constant Symmetry := Symmetries.Trivial'Access;
  Ref: Symmetry renames Symmetries.Reflect;

end Fouriers2;
