with Ada.Text_IO, Ints, Ints.IO, FLts, Rounding, FltStr;
use Ada.Text_IO, Ints, Ints.IO, Flts;
with Ada.Numerics.Generic_Elementary_Functions;

package Long_Floats is

  type Rep is new Flt;

  package Rep_EF is new Ada.Numerics.Generic_Elementary_Functions(Rep);
  package Rep_IO is new Ada.Text_IO.Float_IO(Rep);
  package Rep_FS is new FltStr(Rep);

  function IsZero(R: Rep) return Boolean;
  procedure SetZero(R: in out Rep);
  function Sign(R: Rep) return Integer;
  function Min(R1,R2: Rep) return Rep renames Rep'Min;
  function Max(R1,R2: Rep) return Rep renames Rep'Max;
  procedure Swap(R1,R2: in out Rep);
  function Compare(R1,R2: Rep) return Integer;
  function "*"(F: Flt; R: Rep) return Rep;
  function Convert(R: Rep) return Flt;
  function Approx(R: Rep) return Flt renames Convert;
  function ApproxPut(R: Rep) return FltVector;
  function ApproxGet(V: FltVector) return Rep;
  function Upper(R: Rep) return Flt renames Convert;
  function Lower(R: Rep) return Flt renames Convert;
  function Val(I: Integer) return Rep;
  function Val(F: Flt) return Rep;
  function ValDec(S: String) return Rep renames Rep'Value;
  function ValHex(S: String) return Rep renames Rep_FS.ValHex;
  function DecStr(R: Rep; Digs: Natural := 0) return String renames Rep_FS.DecStr;
  function HexStr(R: Rep; Digs: Natural := 0) return String renames Rep_FS.HexStr;

  function Rounded_Last_Conversion return Boolean;
  procedure Set_Precision(Dummy: in Rep; Binary_Digs: in Positive);
  function Get_Precision(R: Rep) return Positive;
  procedure Set_Rounding_Up(Dummy: in Rep);
  procedure Set_Rounding_Down(Dummy: in Rep);
  procedure Set_Rounding_Nearest(Dummy: in Rep);
  procedure Set_Rounding_Truncate(Dummy: in Rep);
  procedure Set_Rounding_Relaxed(Dummy: in Rep);

  function Sqrt(R: Rep) return Rep renames Rep_EF.Sqrt;
  function Num_Exp(R: Rep) return Rep renames Rep_EF.Exp;
  function Num_Log(R: Rep) return Rep renames Rep_EF.Log;
  function Num_Cos(R: Rep) return Rep renames Rep_EF.Cos;
  function Num_Sin(R: Rep) return Rep renames Rep_EF.Sin;

  pragma Inline_Always (IsZero, SetZero, Swap, Val);

end Long_Floats;
