with Ada.Text_IO, Ada.Command_Line, Ada.Strings.Fixed, Ints.IO, Flts.IO;
use Ada.Text_IO, Ada.Command_Line, Ada.Strings.Fixed, Ints.IO, Flts.IO;

package body Messages is

  use Ints.IO.Int_IO, Flts.IO.Flt_IO;

  Saved_Time: Time := Clock;

  procedure Show(S: in String; Skip: in Natural := 0; NewLine: in Boolean := True) is
  begin
    if Skip <= Verbosity then
      Put(S);
      if NewLine then New_Line; end if;
    end if;
  end Show;

  procedure Show(S: in String; R: in Flt; Skip: in Natural := 0) is
  begin
    if Skip <= Verbosity then
      Put(S);
      Put(R,3,DefaultAft,DefaultExp);
      New_Line;
    end if;
  end Show;

  procedure Show(S: in String; R1,R2: in Flt; Skip: in Natural := 0) is
  begin
    if Skip <= Verbosity then
      Put(S);
      Put(R1,3,DefaultAft,DefaultExp);
      Put(R2,3,DefaultAft,DefaultExp);
      New_Line;
    end if;
  end Show;

  procedure Error(S: in String; Skip: in Natural := 0) is
  begin
    Show(S,Skip);
    if Skip=0 then raise Constraint_Error; end if;
  end Error;

  procedure Error(S: in String; R: in Flt; Skip: in Natural := 0) is
  begin
    Show(S,R,Skip);
    if Skip=0 then raise Constraint_Error; end if;
  end Error;

  procedure Error(S: in String; R1,R2: in Flt; Skip: in Natural := 0) is
  begin
    Show(S,R1,R2,Skip);
    if Skip=0 then raise Constraint_Error; end if;
  end Error;

  procedure TraceEnter(Where: in String) is
  begin
    if Verbosity>1 then
      Put_Line(Head(Dots,2*Indent) & "[ " & Where);
      Indent := Indent+1;
    end if;
  end TraceEnter;

  procedure TraceLeave is
  begin
    if Verbosity>1 then
      Indent := Indent-1;
      Put_Line(Head(Dots,2*Indent) & "]");
    end if;
  end TraceLeave;

  function GetArg return Boolean is
    S: constant String := GetArg;
  begin
    if not (S="True") or (S="False") then
      Error("Messages.GetArg: non-Boolean value");
    end if;
    return S="True";
  end GetArg;

  function GetArg return Integer is
    M,L: Integer;
  begin
    Get(Argument(ArgIndex),M,L);
    ArgIndex := ArgIndex+1;
    return M;
  end GetArg;

  function GetArg return String is
    S: constant String := Argument(ArgIndex);
  begin
    ArgIndex := ArgIndex+1;
    return S;
  end GetArg;

  function GetArg return Flt is
--- careful, input may get rounded
    L: Integer;
    R: Flt;
  begin
    Get(Argument(ArgIndex),R,L);
    ArgIndex := ArgIndex+1;
    return R;
  end GetArg;

  function Age(T: in Timer) return Flt is
--- age in seconds
  begin
    return Flt(To_Duration(Clock-T));
  end Age;

  function Seconds(T: in Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)));
  end Seconds;

  function Minutes(T: in Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)/60.0));
  end Minutes;

  function Hours(T: in Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)/3600.0));
  end Hours;

end Messages;
