with Ada.Finalization, Ada.Strings, Interfaces.C, MP_Arch, Flts;
use Ada.Strings, Interfaces.C, MP_Arch, Flts;

package MPFR.Floats is

  pragma Elaborate_Body;

  type MPFloat is private;

  procedure Set_Precision(Dummy: in MPFloat; Binary_Digits: in Positive);
  function Get_Precision(R: MPFloat) return Positive;
  procedure Set_Rounding_Up(Dummy: in MPFloat);
  procedure Set_Rounding_Down(Dummy: in MPFloat);
  procedure Set_Rounding_Nearest(Dummy: in MPFloat);
  procedure Set_Rounding_Truncate(Dummy: in MPFloat);
  procedure Set_Rounding_Relaxed(Dummy: in MPFloat) renames Set_Rounding_Nearest;

  function IsZero(R: MPFloat) return Boolean;
  procedure SetZero(R: in out MPFloat);
  procedure Swap(R1,R2: in out MPFloat);
  function Compare(R1,R2: MPFloat) return Integer;
  function "<"(R1,R2: MPFloat) return Boolean;
  function "<="(R1,R2: MPFloat) return Boolean;
  function "="(R1,R2: MPFloat) return Boolean;
  function ">="(R1,R2: MPFloat) return Boolean;
  function ">"(R1,R2: MPFloat) return Boolean;

  function Sign(R: MPFloat) return Integer;
  function "abs"(R: MPFloat) return MPFloat;
  function Min(R1,R2: MPFloat) return MPFloat;
  function Max(R1,R2: MPFloat) return MPFloat;
  function "-"(R: MPFloat) return MPFloat;
  function "+"(R1,R2: MPFloat) return MPFloat;
  function "-"(R1,R2: MPFloat) return MPFloat;
  function "*"(R1,R2: MPFloat) return MPFloat;
  function "/"(R1,R2: MPFloat) return MPFloat;
  function "**"(R: MPFloat; N: Integer) return MPFloat;
  function "*"(F: Flt; R: MPFloat) return MPFloat;

  function Upper(R: MPFloat) return Flt;
  function Lower(R: MPFloat) return Flt;
  function Approx(R: MPFloat) return Flt;
  function ApproxPut(R: MPFloat) return FltVector;
  function ApproxGet(V: FltVector) return MPFloat;
  function Val(I: Integer) return MPFloat;
  function Val(F: Flt) return MPFloat;

  function Sqrt(R: MPFloat) return MPFloat;
  function Exp(R: MPFloat) return MPFloat;
  function Num_Exp(R: MPFloat) return MPFloat renames Exp;
  function Log(R: MPFloat) return MPFloat;
  function Num_Log(R: MPFloat) return MPFloat renames Log;
  function Cos(R: MPFloat) return MPFloat;
  function Num_Cos(R: MPFloat) return MPFloat renames Cos;
  function Sin(R: MPFloat) return MPFloat;
  function Num_Sin(R: MPFloat) return MPFloat renames Sin;

  function Image(R: MPFloat; Base: Positive; Digs: Natural := 0) return String;
  function Value(S: String; Base: Positive) return MPFloat;
  function ValDec(S: String) return MPFloat;
  function ValHex(S: String) return MPFloat;
  function DecStr(R: MPFloat; Digs: Natural := 0) return String;
  function HexStr(R: MPFloat; Digs: Natural := 0) return String;
  function Rounded_Last_Conversion return Boolean;

  MPFR_Exception: exception;

private

  subtype AFC is Ada.Finalization.Controlled;

  type MPFloat is new AFC with
    record
      Val: MPFR_T;
    end record;

  overriding procedure Initialize(R: in out MPFloat);
  overriding procedure Adjust(R: in out MPFloat);
  overriding procedure Finalize(R: in out MPFloat);

  function Problem return Boolean;
  procedure Report(Name: in String);
  procedure Clear_Flags;
  function Raised_Inexact return Boolean;
  function NDigits(P: MP_Prec_T; Base: Positive) return Size_T;
  function EffLast(F: Positive; S: String) return Positive;

  Precision: MP_Prec_T := Long_Double'Machine_Mantissa;
  Rounding: MPFR_Rnd_T := GMP_RNDN;
  Just_Rounded: Boolean := True;

end MPFR.Floats;
