with Interfaces.C, Interfaces.C.Strings, MP_Arch;
use Interfaces.C, Interfaces.C.Strings, MP_Arch;

package MPFR is    --- see also mp.h, gmp.h, mpfr.h

  subtype MPFR_Base is Int range   2 .. 36;
  type MP_Exp_T_Ptr is access MP_Exp_T;

  type MPFR_T is private;

  type MPFR_Rnd_T is (GMP_RNDN, -- round to nearest
                      GMP_RNDZ, -- round toward zero
                      GMP_RNDU, -- round up
                      GMP_RNDD, -- round down
                      GMP_RND_MAX);

  for MPFR_Rnd_T use (GMP_RNDN    => 0,
                      GMP_RNDZ    => 1,
                      GMP_RNDU    => 2,
                      GMP_RNDD    => 3,
                      GMP_RND_MAX => 4 );

  procedure MPFR_Init2(R: in out MPFR_T; Prec: in MP_Prec_T);
  procedure MPFR_Clear(R: in out MPFR_T);

  procedure MPFR_Set_Default_Prec(Prec: in MP_Prec_T);
  function  MPFR_Get_Default_Prec return MP_Prec_T;
  procedure MPFR_Set_Prec(R: in out MPFR_T; Prec: in MP_Prec_T);
  function MPFR_Get_Prec(R: in MPFR_T) return MP_Prec_T;

  procedure MPFR_Set(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Set_Si(ROp: in out MPFR_T; Op: in Long; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Set_Ld(ROp: in out MPFR_T; Op: in Long_Double; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Init_Set_Str(R: in out MPFR_T; C: in Chars_Ptr; Base: in MPFR_Base; Rnd: in MPFR_Rnd_T);
  function MPFR_Get_Str(C: in Chars_Ptr; ExpPtr: MP_Exp_T_Ptr; Base: in MPFR_Base; NDigits: in Size_T;
                        Op: in MPFR_T; Rnd: in MPFR_Rnd_T) return Chars_Ptr;
  function MPFR_Get_Ld(Op: in MPFR_T; Rnd: in MPFR_Rnd_T) return Long_Double;
  procedure MPFR_Free_Str(C: in Chars_Ptr);
  procedure MPFR_Swap(Op1,Op2: in out MPFR_T);
  function MPFR_Cmp(Op1, Op2: in MPFR_T) return Int;

  function MPFR_Sgn(Op : MPFR_T) return Int;
  procedure MPFR_Abs(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Neg(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Add(ROp: in out MPFR_T; Op1, Op2: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Sub(ROp: in out MPFR_T; Op1, Op2: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Mul(ROp: in out MPFR_T; Op1, Op2: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Pow_Si(ROp: in out MPFR_T; Op1: in MPFR_T; Op2: in Long; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Div(ROp: in out MPFR_T; Op1, Op2: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Sqrt(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Exp(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Log(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Cos(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);
  procedure MPFR_Sin(ROp: in out MPFR_T; Op: in MPFR_T; Rnd: in MPFR_Rnd_T);

  procedure MPFR_Clear_Flags;
  function MPFR_Underflow_P return Int;
  function MPFR_Overflow_P return Int;
  function MPFR_NanFlag_P return Int;
  function MPFR_InexFlag_P return Int;
  function MPFR_ERangeFlag_P return Int;

private

  Pragma Convention(C, MPFR_Rnd_T);

  type MP_Limb_T_Ptr is access all MP_Limb_T;  -- MP_Limb_A

  type MPFR_Struct is
    record
      MP_Prec: MP_Prec_T;
      MPFR_Sign: MP_Sign_T;
      MPFR_Exp : MP_Exp_T;
      MPFR_D   : MP_Limb_T_Ptr;
    end record;

  type MPFR_T   is array(1 .. 1) of MPFR_Struct;
  type MPFR_Ptr is access all MPFR_Struct;

  Pragma Convention (C, MPFR_Struct);
  Pragma Convention (C, MPFR_T);
  Pragma Convention (C, MPFR_Ptr);

  pragma Import (C, MPFR_Init2, "mpfr_init2");
  pragma Import (C, MPFR_Clear, "mpfr_clear");

  pragma Import (C, MPFR_Set_Default_Prec, "mpfr_set_default_prec");
  pragma Import (C, MPFR_Get_Default_Prec, "mpfr_get_default_prec");
  pragma Import (C, MPFR_Set_Prec, "mpfr_set_prec");
  pragma Import (C, MPFR_Get_Prec, "mpfr_get_prec");

  pragma Import (C, MPFR_Set, "mpfr_set");
  pragma Import (C, MPFR_Set_Si, "mpfr_set_si");
  pragma Import (C, MPFR_Set_Ld, "mpfr_set_ld");
  pragma Import (C, MPFR_Init_Set_Str, "mpfr_init_set_str");
  pragma Import (C, MPFR_Get_Str, "mpfr_get_str");
  pragma Import (C, MPFR_Get_Ld, "mpfr_get_ld");
  pragma Import (C, MPFR_Free_Str, "mpfr_free_str");
  pragma Import (C, MPFR_Swap, "mpfr_swap");
  pragma Import (C, MPFR_Cmp, "mpfr_cmp");

  pragma Import (C, MPFR_Sgn, "mpfr_sgn");
  pragma Import (C, MPFR_Abs, "mpfr_abs");
  pragma Import (C, MPFR_Neg, "mpfr_neg");
  pragma Import (C, MPFR_Add, "mpfr_add");
  pragma Import (C, MPFR_Sub, "mpfr_sub");
  pragma Import (C, MPFR_Mul, "mpfr_mul");
  pragma Import (C, MPFR_Pow_Si, "mpfr_pow_si");
  pragma Import (C, MPFR_Div, "mpfr_div");
  pragma Import (C, MPFR_Sqrt, "mpfr_sqrt");
  pragma Import (C, MPFR_Exp, "mpfr_exp");
  pragma Import (C, MPFR_Log, "mpfr_log");
  pragma Import (C, MPFR_Cos, "mpfr_cos");
  pragma Import (C, MPFR_Sin, "mpfr_sin");

  pragma Import (C, MPFR_Clear_Flags, "mpfr_clear_flags");
  pragma Import (C, MPFR_Underflow_P, "mpfr_underflow_p");
  pragma Import (C, MPFR_Overflow_P, "mpfr_overflow_p");
  pragma Import (C, MPFR_NanFlag_P, "mpfr_nanflag_p");
  pragma Import (C, MPFR_InexFlag_P, "mpfr_inexflag_p");
  pragma Import (C, MPFR_ERangeFlag_P, "mpfr_erangeflag_p");

end MPFR;
