with Messages, Ada.Strings.Unbounded.Text_IO;
use Messages;

package body Numerics is

  RepZero:  constant Rep  := Val(Zero);
  RepOne:   constant Rep  := Val(One);
  RepThree: constant Rep  := Val(Three);

  function ReturnTrue(S: Numeric) return Boolean is
  begin
    return True;
  end ReturnTrue;

  function ReturnZero(S: Numeric) return Numeric is
    O: constant Rep := RepZero;
  begin
    return Numeric(O);
  end ReturnZero;

  function ReturnZero(R: Flt) return Numeric is
    O: constant Rep := RepZero;
  begin
    return Numeric(O);
  end ReturnZero;

  function Average(S,T: Numeric) return Numeric is
  begin
    return Numeric(Half*(Rep(S)+Rep(T)));
  end Average;

------------------------

  function IsZero(S: Numeric) return Boolean is
  begin
    return IsZero(Rep(S));
  end IsZero;

  procedure SetZero(S: in out Numeric) is
  begin
    SetZero(Rep(S));
  end SetZero;

  function Copy(S: Numeric) return Numeric is
    T: constant Numeric := S;
  begin
    return T;
  end Copy;

  procedure Enclosure(S: in out Numeric; Binary_Digs: in Positive) is
  begin
    if Binary_Digs<Precision then
      Set_Precision(RepZero,Binary_Digs);
      declare
        T: constant Numeric := S;
      begin
        Set_Precision(RepZero,Precision);
        S := T;
      end;
    end if;
  end Enclosure;

  function Approx(S: Numeric) return Flt is
  begin
    return Approx(Rep(S));
  end Approx;

  function FPut(R: Numeric) return FltVector is
  begin
    return ApproxPut(Rep(R));
  end FPut;

  function FGet(V: FltVector) return Numeric is
  begin
    return Numeric(ApproxGet(V));
  end FGet;

  function Scal(K: Integer) return Numeric is
  begin
    return Numeric(Val(K));
  end Scal;

  function Scal(R: Flt) return Numeric is
  begin
    return Numeric(Val(R));
  end Scal;

  function Scal(R1,R2: Flt) return Numeric is
  begin
    return Numeric(Half*(Val(R1)+Val(R2)));
  end Scal;

  procedure Swap(S1,S2: in out Numeric) is
  begin
    Swap(Rep(S1),Rep(S2));
  end Swap;

--------------------

  function "="(S1,S2: Numeric) return Boolean is
  begin
    return (Rep(S1)=Rep(S2));
  end "=";

  function Disjoint(S1,S2: Numeric) return Boolean is
  begin
    return not (Rep(S1)=Rep(S2));
  end Disjoint;

  function "<"(S1,S2: Numeric) return Boolean is
  begin
    return (Rep(S1)<Rep(S2));
  end "<";

  function "<="(S1,S2: Numeric) return Boolean is
  begin
    return (Rep(S1) <= Rep(S2));
  end "<=";

  function ">"(S1,S2: Numeric) return Boolean is
  begin
    return (Rep(S1)>Rep(S2));
  end ">";

  function ">="(S1,S2: Numeric) return Boolean is
  begin
    return (Rep(S1) >= Rep(S2));
  end ">=";

  function Min(S1,S2: Numeric) return Numeric is
  begin
    return Numeric(Min(Rep(S1),Rep(S2)));
  end Min;

  function Max(S1,S2: Numeric) return Numeric is
  begin
    return Numeric(Max(Rep(S1),Rep(S2)));
  end Max;

--------------------

  function "abs"(S: Numeric) return Numeric is
  begin
    return Numeric(abs(Rep(S)));
  end "abs";

  function MaxNorm(S: Numeric) return Flt is
  begin
    return Approx(abs(Rep(S)));
  end MaxNorm;

  procedure Neg(S: in out Numeric) is
  begin
    S := Numeric(-Rep(S));
  end Neg;

  function "-"(S: Numeric) return Numeric is
  begin
    return Numeric(-Rep(S));
  end "-";

  procedure Add(S: in Numeric; T: in out Numeric) is
  begin
    T := Numeric(Rep(T)+Rep(S));
  end Add;

  function "+"(S,T: Numeric) return Numeric is
  begin
    return Numeric(Rep(S)+Rep(T));
  end "+";

  procedure Sub(S: in Numeric; T: in out Numeric) is
  begin
    T := Numeric(Rep(T)-Rep(S));
  end Sub;

  function "-"(S,T: Numeric) return Numeric is
  begin
    return Numeric(Rep(S)-Rep(T));
  end "-";

  procedure Mult(R: Flt; S: in out Numeric) is
  begin
    S := Numeric(R*Rep(S));
  end Mult;

  function "*"(R: Flt; S: Numeric) return Numeric is
  begin
    return Numeric(R*Rep(S));
  end "*";

  procedure Mult(R: in Rep; S: in out Numeric) is
  begin
    S := Numeric(R*Rep(S));
  end Mult;

  function "*"(R: Rep; S: Numeric) return Numeric is
  begin
    return Numeric(R*Rep(S));
  end "*";

  function Sqr(S: Numeric) return Numeric is
  begin
    return Numeric(Rep(S)*Rep(S));
  end Sqr;

  procedure Mult(S: in Numeric; T: in out Numeric) is
  begin
    T := Numeric(Rep(S)*Rep(T));
  end Mult;

  function "*"(S,T: Numeric) return Numeric is
  begin
    return Numeric(Rep(S)*Rep(T));
  end "*";

  procedure AddProd(S1,S2: in Numeric; S3: in out Numeric) is
  begin
    S3 := Numeric(Rep(S3)+Rep(S1)*Rep(S2));
  end AddProd;

  function Inv(S: Numeric) return Numeric is
  begin
    return Numeric(RepOne/Rep(S));
  end Inv;

  function "/"(S,T: Numeric) return Numeric is
  begin
    return Numeric(Rep(S)/Rep(T));
  end "/";

  function "**"(S: Numeric; I: Integer) return Numeric is
  begin
    return Numeric(Rep(S)**I);
  end "**";

---------------------------

  function Sqrt(S: Numeric) return Numeric is
  begin
    return Numeric(Sqrt(Rep(S)));
  end Sqrt;

  function Exp(S: Numeric) return Numeric is
  begin
    return Numeric(Num_Exp(Rep(S)));
  end Exp;

  function Log(S: Numeric) return Numeric is
  begin
    return Numeric(Num_Log(Rep(S)));
  end Log;

  function Cos(S: Numeric) return Numeric is
  begin
    return Numeric(Num_Cos(Rep(S)));
  end Cos;

  function Sin(S: Numeric) return Numeric is
  begin
    return Numeric(Num_Sin(Rep(S)));
  end Sin;

  function Pi return Numeric is
--- Bailey, 1997
    Iter: constant Integer := 20;
    S: constant Numeric := Inv(Scal(16));
    P: Numeric := Scal(0);
  begin
    for K in reverse 0 .. Iter loop
      P := S*P+(Inv(Scal(8*K+1))-Inv(Scal(16*K+8))-Inv(Scal(32*K+20))-Inv(Scal(32*K+24)));
    end loop;
    return Scal(4)*P;
  end Pi;

----------------------------

  procedure ShowType(S: in Numeric) is
  begin
    Put_Line("Scalar type is Numeric");
  end ShowType;

  procedure Show1(N: in String; S: in Numeric; NewLine: in Boolean := True) is
    Digs: constant Natural := 0; -- use default
  begin
    Put_Line(N & DecStr(Rep(S),Digs));
    if NewLine then New_Line; end if;
  end Show1;

  procedure Show2(N: in String; S1,S2: in Numeric; NewLine: in Boolean := True) is
    Digs: constant Natural := 0; -- use default
  begin
    Put_Line(N & DecStr(Rep(S1),Digs) & "   " & DecStr(Rep(S2),Digs));
    if NewLine then New_Line; end if;
  end Show2;

  procedure Put(F: in File_Type; S: in Numeric; Decimal: in Boolean := False) is
  begin
    if Decimal then
      Put_Line(F,DecStr(Rep(S)));
    else
      Put_Line(F,HexStr(Rep(S)));
    end if;
  end Put;

  procedure Get(F: in File_Type; S: in out Numeric; Decimal: in Boolean := False) is
    use Ada.Strings.Unbounded;
    U: Unbounded_String := Null_Unbounded_String;
  begin
    while Length(U)=0 loop
      Text_IO.Get_Line(F,U);
    end loop;
    if Decimal then
      S := Numeric(ValDec(To_String(U)));
    else
      S := Numeric(ValHex(To_String(U)));
    end if;
  end Get;

  procedure Write(Name: in String; S: in Numeric; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Writing " & Name);
    Create(F,Out_File,Name);
    Put(F,S,Decimal);
    Close(F);
  end Write;

  procedure Read(Name: in String; S: in out Numeric; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Reading " & Name);
    Open(F,In_File,Name);
    Get(F,S,Decimal);
    Close(F);
  end Read;

  procedure Set_Active_Mode(Dummy: in Numeric) is
  begin
    Show("numerics active",2);
    Active := True;
    Set_Precision(RepZero,Precision);
    Set_Rounding_Relaxed(RepZero);
  end Set_Active_Mode;

  procedure Set_Wait_Mode(Dummy: in Numeric) is
  begin
    Show("numerics waiting",2);
    Active := False;
  end Set_Wait_Mode;

  procedure Set_New_Precision(Dummy: in Numeric; Binary_Digs: in Positive) is
  begin
    Precision := Binary_Digs;
    if Active then
      Set_Precision(RepZero,Precision);
    end if;
  end Set_New_Precision;

  function Get_Precision(S: Numeric) return Positive is
  begin
    return Get_Precision(Rep(S));
  end Get_Precision;

end Numerics;
