with Ada.Text_IO, Ints, Ints.IO, Flts;
use Ada.Text_IO, Ints, Ints.IO, Flts;

generic

-------------------- begin_include reps Rep

  type Rep is private;  --- used to define intervals, balls, ...

  with function IsZero(R: Rep) return Boolean is <>;
  with procedure SetZero(R: in out Rep) is <>;
  with function Sign(R: Rep) return Integer is <>;
  with function "abs"(R: Rep) return Rep is <>;
  with function Min(R1,R2: Rep) return Rep is <>;
  with function Max(R1,R2: Rep) return Rep is <>;
  with procedure Swap(R1,R2: in out Rep) is <>;

  with function Compare(R1,R2: Rep) return Integer is <>;
  with function "<"(R1,R2: Rep) return Boolean is <>;
  with function "<="(R1,R2: Rep) return Boolean is <>;
  with function "="(R1,R2: Rep) return Boolean is <>;
  with function ">="(R1,R2: Rep) return Boolean is <>;
  with function ">"(R1,R2: Rep) return Boolean is <>;

  with function "-"(R: Rep) return Rep is <>;
  with function "+"(R1,R2: Rep) return Rep is <>;
  with function "-"(R1,R2: Rep) return Rep is <>;
  with function "*"(R1,R2: Rep) return Rep is <>;
  with function "/"(R1,R2: Rep) return Rep is <>;
  with function "**"(R: Rep; N: Integer) return Rep is <>;
  with function "*"(F: Flt; R: Rep) return Rep is <>;
  with function Sqrt(R: Rep) return Rep is <>;

  with function Upper(R: Rep) return Flt is <>;
  with function Lower(R: Rep) return Flt is <>;
  with function Approx(R: Rep) return Flt is <>;
  with function ApproxPut(R: Rep) return FltVector is <>;
  with function ApproxGet(V: FltVector) return Rep is <>;
  with function Val(I: Integer) return Rep is <>;
  with function Val(F: Flt) return Rep is <>;
  with function ValDec(S: String) return Rep is <>;
  with function ValHex(S: String) return Rep is <>;
  with function DecStr(R: Rep; Digs: Natural := 0) return String is <>; -- default, if NDig=0
  with function HexStr(R: Rep; Digs: Natural := 0) return String is <>; -- full, if NDig=0
  with function Rounded_Last_Conversion return Boolean is <>;

  with procedure Set_Precision(Dummy: in Rep; Binary_Digs: in Positive) is <>;
  with function Get_Precision(R: Rep) return Positive is <>;
  with procedure Set_Rounding_Up(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Down(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Nearest(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Truncate(Dummy: in Rep) is <>;
  with procedure Set_Rounding_Relaxed(Dummy: in Rep) is <>;

--- approximate
  with function Num_Exp(R: Rep) return Rep is <>;
  with function Num_Log(R: Rep) return Rep is <>;
  with function Num_Cos(R: Rep) return Rep is <>;
  with function Num_Sin(R: Rep) return Rep is <>;

-------------------- end_include reps Rep

package Numerics is

  type Numeric is new Rep;

  function ReturnTrue(S: Numeric) return Boolean;
  function ReturnZero(S: Numeric) return Numeric;
  function ReturnZero(R: Flt) return Numeric;
  function Average(S,T: Numeric) return Numeric;

--- basic
  function "="(S1,S2: Numeric) return Boolean;
  function IsNumeric(S: Numeric) return Boolean renames ReturnTrue;
  function IsZero(S: Numeric) return Boolean;
  procedure SetZero(S: in out Numeric);
  function Copy(S: Numeric) return Numeric;
  procedure Enclosure(S: in out Numeric; Binary_Digs: in Positive);
  function Approx(S: Numeric) return Flt;
  function FPut(R: Numeric) return FltVector;
  function FGet(V: FltVector) return Numeric;
  function Scal(K: Integer) return Numeric;
  function Scal(R: Flt) return Numeric;
  function Scal(R1,R2: Flt) return Numeric;
  procedure Swap(S1,S2: in out Numeric);

--- intervals
  function Sup(S: Numeric) return Flt renames Approx;
  function Inf(S: Numeric) return Flt renames Approx;
  function Disjoint(S1,S2: Numeric) return Boolean;
  function "<"(S1,S2: Numeric) return Boolean;
  function "<="(S1,S2: Numeric) return Boolean;
  function ">"(S1,S2: Numeric) return Boolean;
  function ">="(S1,S2: Numeric) return Boolean;
  function Min(S1,S2: Numeric) return Numeric;
  function Max(S1,S2: Numeric) return Numeric;

--- sets
  function Ball0(S: Numeric) return Numeric renames ReturnZero;
  function Ball0(R: Flt) return Numeric renames ReturnZero;
  function Center(S: Numeric) return Numeric renames Copy;
  procedure ResetCenter(S: in out Numeric) renames SetZero;
  function Subset(S1,S2: Numeric) return Boolean renames "=";
  function Union(S1,S2: Numeric) return Numeric renames Average;
  function Intersection(S1,S2: Numeric) return Numeric renames Average;

--- ops
  function "abs"(S: Numeric) return Numeric;
  function Norm(S: Numeric) return Numeric renames "abs";
  function MaxNorm(S: Numeric) return Flt;
  procedure Neg(S: in out Numeric);
  function "-"(S: Numeric) return Numeric;
  procedure Add(S: in Numeric; T: in out Numeric);
  function "+"(S,T: Numeric) return Numeric;
  procedure Sub(S: in Numeric; T: in out Numeric);
  function "-"(S,T: Numeric) return Numeric;
  procedure Mult(R: Flt; S: in out Numeric);
  function "*"(R: Flt; S: Numeric) return Numeric;
  function Sqr(S: Numeric) return Numeric;
  procedure Mult(S: in Numeric; T: in out Numeric);
  function "*"(S,T: Numeric) return Numeric;
  procedure AddProd(S1,S2: in Numeric; S3: in out Numeric);
  function Inv(S: Numeric) return Numeric;
  function "/"(S,T: Numeric) return Numeric;
  function "**"(S: Numeric; I: Integer) return Numeric;

--- fun
  function Sqrt(S: Numeric) return Numeric;
  function Exp(S: Numeric) return Numeric;
  function Cos(S: Numeric) return Numeric;
  function Sin(S: Numeric) return Numeric;
  function Log(S: Numeric) return Numeric;
  function Pi return Numeric;

--- io
  procedure ShowType(S: in Numeric);
  procedure Show1(N: in String; S: in Numeric; NewLine: in Boolean := True);
  procedure Show2(N: in String; S1,S2: in Numeric; NewLine: in Boolean := True);
  procedure Put(F: in File_Type; S: in Numeric; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; S: in out Numeric; Decimal: in Boolean := False);
  procedure Write(Name: in String; S: in Numeric; Decimal: in Boolean := False);
  procedure Read(Name: in String; S: in out Numeric; Decimal: in Boolean := False);

--- misc
  procedure Set_Active_Mode(Dummy: in Numeric);
  procedure Set_Wait_Mode(Dummy: in Numeric);
  procedure Set_New_Precision(Dummy: in Numeric; Binary_Digs: in Positive);
  function Get_Precision(S: Numeric) return Positive;

  pragma Inline (Copy, Scal, Neg, Add, Sub, Mult, AddProd, Inv);

private

  Active: Boolean := True;
  Precision: Positive := Get_Precision(Val(0));

end Numerics;
