with Flts, Numerics, Constants, Vectors, Vectors.Ops, Vectors.Num, Fouriers2;
use Flts;

pragma Elaborate_All (Numerics, Constants, Vectors, Vectors.Ops, Vectors.Num, Fouriers2);

generic

  Deg1,Deg2: in Positive;

-------------------- begin_include reps NRep

  type NRep is private;  --- used to define intervals, balls, ...

  with function IsZero(R: NRep) return Boolean is <>;
  with procedure SetZero(R: in out NRep) is <>;
  with function Sign(R: NRep) return Integer is <>;
  with function "abs"(R: NRep) return NRep is <>;
  with function Min(R1,R2: NRep) return NRep is <>;
  with function Max(R1,R2: NRep) return NRep is <>;
  with procedure Swap(R1,R2: in out NRep) is <>;

  with function Compare(R1,R2: NRep) return Integer is <>;
  with function "<"(R1,R2: NRep) return Boolean is <>;
  with function "<="(R1,R2: NRep) return Boolean is <>;
  with function "="(R1,R2: NRep) return Boolean is <>;
  with function ">="(R1,R2: NRep) return Boolean is <>;
  with function ">"(R1,R2: NRep) return Boolean is <>;

  with function "-"(R: NRep) return NRep is <>;
  with function "+"(R1,R2: NRep) return NRep is <>;
  with function "-"(R1,R2: NRep) return NRep is <>;
  with function "*"(R1,R2: NRep) return NRep is <>;
  with function "/"(R1,R2: NRep) return NRep is <>;
  with function "**"(R: NRep; N: Integer) return NRep is <>;
  with function "*"(F: Flt; R: NRep) return NRep is <>;
  with function Sqrt(R: NRep) return NRep is <>;

  with function Upper(R: NRep) return Flt is <>;
  with function Lower(R: NRep) return Flt is <>;
  with function Approx(R: NRep) return Flt is <>;
  with function ApproxPut(R: NRep) return FltVector is <>;
  with function ApproxGet(V: FltVector) return NRep is <>;
  with function Val(I: Integer) return NRep is <>;
  with function Val(F: Flt) return NRep is <>;
  with function ValDec(S: String) return NRep is <>;
  with function ValHex(S: String) return NRep is <>;
  with function DecStr(R: NRep; Digs: Natural := 0) return String is <>; -- default, if NDig=0
  with function HexStr(R: NRep; Digs: Natural := 0) return String is <>; -- full, if NDig=0
  with function Rounded_Last_Conversion return Boolean is <>;

  with procedure Set_Precision(Dummy: in NRep; Binary_Digs: in Positive) is <>;
  with function Get_Precision(R: NRep) return Positive is <>;
  with procedure Set_Rounding_Up(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Down(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Nearest(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Truncate(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Relaxed(Dummy: in NRep) is <>;

--- approximate
  with function Num_Exp(R: NRep) return NRep is <>;
  with function Num_Log(R: NRep) return NRep is <>;
  with function Num_Cos(R: NRep) return NRep is <>;
  with function Num_Sin(R: NRep) return NRep is <>;

-------------------- end_include reps NRep

package NumFouriers2 is

  package Numbers is new Numerics (Rep => NRep);
  use Numbers;

  package NConstants is new Constants (Scalar => Numeric);
  package NVectors is new Vectors (Scalar => Numeric);
  package NVectors_Ops is new NVectors.Ops;
  package NVectors_Num is new NVectors.Num;
  package NFouriers2 is new Fouriers2 (Deg1 => Deg1, Deg2 => Deg2, Scalar => Numeric,
          SC => NConstants, SV => NVectors, SV_Ops => NVectors_Ops, SV_Num => NVectors_Num);

  subtype Numeric is Numbers.Numeric;
  subtype NVector is NFouriers2.ScalVec;
  subtype NDiag is NFouriers2.ScalDiag;
  subtype NMatrix is NFouriers2.ScalMat;
  subtype NSimDiag is NFouriers2.ScalSimDiag;
  subtype NLU is NFouriers2.ScalLU;
  subtype NFourier2 is NFouriers2.Fourier2;
  subtype NParam is NFouriers2.Param;
  subtype NMap is NFouriers2.Map;
  subtype NDMap is NFouriers2.DMap;

  function N2R(N: Numeric) return FltVector renames Numbers.FPut;
  function R2N(R: FltVector) return Numeric renames Numbers.FGet;

end NumFouriers2;
