with Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO, Symmetries, Numerics, Square;
use Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO, Symmetries;

with Long_Floats; use Long_Floats;

procedure PlotBifFun is

  type NRep is new Rep;  -- rep used for numeric computations
  type SRep is new Rep;  -- rep used to define Scalar

  package Scalars is new Numerics (Rep => SRep);
  use Scalars;
  type Scalar is new Numeric;

  Deg: constant Positive := GetArg;

  package SQ is new Square (Deg => Deg, NRep => NRep, Scalar => Scalar);
  use SQ;
  use SConstants, SVectors, SFouriers2, SContr;
  subtype Fun is SFourier2;

  Nop: constant Symmetry := Symmetries.Trivial'Access;
  Ref: Symmetry renames Symmetries.Reflect;

  LSize:  constant Integer := 150;         -- used for eigenvalues/vectors
  Eps:    constant Radius  := 1.0E-20;     -- for approx solutions
  Rad:    constant Radius  := Half**60;    -- domain
  Lambda: constant Flt     := GetArg;
  P:      constant SParam2 := (Scal(Lambda),ScalZero);

  procedure ShowParam is
  begin
    Show_Param;
    Show1(" LSize =",LSize);
    Show1("   Eps =",Eps);
    Show1("Lambda =",Lambda);
    Show1("   Rad =",Rad);
  end ShowParam;

  NPlot: constant Integer := 10;
  Mu0: Flt;
  E: Scalar;
  F,U: Fun;

begin
  Verbosity := 0;
  ShowParam;
  Ref := Symmetries.Trivial'Access;     -- no symmetry

  Read("nfix-0.09532152-25.data",F);
  ApproxEigen1(DMapA'Access,P,F,E,U,LSize);
  Write("eigen.data",U);
  Mu0 := Approx(Compo1(F,U));
  BifFunPlotA(0.09,0.10,Mu0-0.02,Mu0+0.02,U,F,NPlot);

end PlotBifFun;
