
package Rounding is          -- for x87 floating point processor
                             -- 80 bit extended format (64 bit mantissa)
                             -- gnat specific code, but easy to adapt
  pragma Elaborate_Body;

  type Rounding_Mode is private;

  procedure Set_Rounding_Up;
  procedure Set_Rounding_Down;
  procedure Set_Rounding_Nearest;
  procedure Set_Rounding_Truncate;
  procedure Set_Rounding_Relaxed;
  function Get_Rounding_Mode return Rounding_Mode;

private

  type Rounding_Mode is mod 2**16;

  procedure Load_Control_Word(W: in Rounding_Mode);

  Nearest:  constant Rounding_Mode := 2#0011_0110_0000#;
  Down:     constant Rounding_Mode := 2#0111_0110_0000#;
  Up:       constant Rounding_Mode := 2#1011_0110_0000#;
  Truncate: constant Rounding_Mode := 2#1111_0110_0000#;
  Relaxed:  constant Rounding_Mode := 2#0011_0111_0010#;
--                                      abcd   ef ghij
--  ab=00  round toward zero
--  ab=10  round up
--  cd=11  64 bit mantissa
--   e=1   allow rounding (mask exception)
--   f=0   disallow underflows
--   g=0   overflows not masked     (exception instead of NaN)
--   h=0   division by 0 not masked (exception instead of infinity)
--   i=0   disallow denormalized numbers
--   j=0   invalid operations not masked

  Current_Rounding_Mode: Rounding_Mode;

end Rounding;
