with Ints.IO, Flts.IO, Messages;
use Ints.IO, Flts.IO, Messages;

package body Square is

  use NVectors, NVectors_Ops, NVectors_Num;

  function N2S(N: Numeric) return Scalar is
  begin
    return R2S(N2R(N));
  end N2S;

  function S2N(S: Scalar) return Numeric is
  begin
    return R2N(S2R(S));
  end S2N;

  procedure N2S(V: in NVector; W: in out SVector) is
  begin
    for K in V'Range loop
      W(K) := R2S(N2R(V(K)));
    end loop;
  end N2S;

  procedure S2N(V: in SVector; W: in out NVector) is
  begin
    for K in V'Range loop
      W(K) := R2N(S2R(V(K)));
    end loop;
  end S2N;

  function S2N(V: SVector) return NVector is
    W: NVector(V'Range);
  begin
    S2N(V,W);
    return W;
  end S2N;

  procedure N2S(A: in NMatrix; B: in out SMatrix) is
  begin
    for M in A'Range(1) loop
      for N in A'Range(2) loop
        B(M,N) := R2S(N2R(A(M,N)));
      end loop;
    end loop;
  end N2S;

  procedure S2N(A: in SMatrix; B: in out NMatrix) is
  begin
    for M in A'Range(1) loop
      for N in A'Range(2) loop
        B(M,N) := R2N(S2R(A(M,N)));
      end loop;
    end loop;
  end S2N;

  function S2N(A: SMatrix) return NMatrix is
    B: NMatrix(A'Range(1),A'Range(2));
  begin
    S2N(A,B);
    return B;
  end S2N;

  procedure N2S(F: in NFourier2; G: in out SFourier2) is
  begin
    G.T.P := F.T.P;
    G.T.R := F.T.R;
    N2S(F.C,G.C);
    G.E := SFouriers2.Zero_FErr;
  end N2S;

  procedure S2N(F: in SFourier2; G: in out NFourier2) is
  begin
    G.T.P := F.T.P;
    G.T.R := F.T.R;
    S2N(F.C,G.C);
    G.E := NFouriers2.Zero_FErr;
  end S2N;

  function S2N(F: SFourier2) return NFourier2 is
    G: NFourier2;
  begin
    S2N(F,G);
    return G;
  end S2N;

  function S2N(Y: SMap) return NMap is
    use SFouriers2.Contr,NFouriers2.Contr;
  begin
    if Y=SFouriers2.MapA'Access then return NFouriers2.MapA'Access; end if;
    return null;
  end S2N;

  function S2N(Y: SDMap) return NDMap is
    use SFouriers2.Contr,NFouriers2.Contr;
  begin
    if Y=SFouriers2.DMapA'Access then return NFouriers2.DMapA'Access; end if;
    return null;
  end S2N;

-------------------------

  procedure Show_Param is
  begin
    ShowType(ScalZero);
    Show1("S-Bits =",Get_Scalar_Precision);
    Show1("N-Bits =",Get_Numeric_Precision);
    Show1("   Deg =",Deg);
  end Show_Param;

  procedure Set_Scalar_Precision(Binary_Digs: in Positive) is
  begin
    Set_New_Precision(SDummy,Binary_Digs);
  end Set_Scalar_Precision;

  procedure Set_Numeric_Precision(Binary_Digs: in Positive) is
  begin
    Numbers.Set_New_Precision(NDummy,Binary_Digs);
  end Set_Numeric_Precision;

  procedure Set_Precision(Binary_Digs: in Positive) is
  begin
    Set_Scalar_Precision(Binary_Digs);
    Set_Numeric_Precision(Binary_Digs);
  end Set_Precision;

  function Get_Scalar_Precision return Positive is
  begin
    return  Get_Precision(Scal(0));
  end Get_Scalar_Precision;

  function Get_Numeric_Precision return Positive is
  begin
    return Numbers.Get_Precision(Numbers.Scal(0));
  end Get_Numeric_Precision;

  procedure Enter_Numeric_Mode is
  begin
    Set_Wait_Mode(SDummy);
    Numbers.Set_Active_Mode(NDummy);
  end Enter_Numeric_Mode;

  procedure Restore_Scalar_Mode is
  begin
    Numbers.Set_Wait_Mode(NDummy);
    Set_Active_Mode(SDummy);
  end Restore_Scalar_Mode;

-------------------------

  procedure ApproxEigen(A: in SMatrix; Wr,Wi: in out SVector; Vr,Vi: in out SMatrix) is
  begin
    TraceEnter("Square.ApproxEigen");
    Enter_Numeric_Mode;
    if Trunc then
      SVectors_Num.Eigen(A,Wr,Wi,Vr,Vi);
    else
      declare
        NWr,NWi: NVector(Wr'Range);
        NVr,NVi: Nmatrix(Vr'Range(1),Vr'Range(2));
      begin
        NVectors_Num.Eigen(S2N(A),NWr,NWi,NVr,NVi);
        N2S(NWr,Wr);
        N2S(NWi,Wi);
        N2S(NVr,Vr);
        N2S(NVi,Vi);
      end;
    end if;
    Restore_Scalar_Mode;
    TraceLeave;
  end ApproxEigen;

  procedure ApproxEigen(A: in SMatrix; Er,Ei: in out Scalar; Vr,Vi: in out SVector) is
  begin
    TraceEnter("Square.ApproxEigen");
    Enter_Numeric_Mode;
    if Trunc then
      SVectors_Num.ClosestEigen(A,Er,Ei,Vr,Vi);
    else
      declare
        NEr: Numeric := S2N(Er);
        NEi: Numeric := S2N(Ei);
        NVr,NVi: NVector(Vr'Range);
      begin
        NVectors_Num.ClosestEigen(S2N(A),NEr,NEi,NVr,NVi);
        Er := N2S(NEr);
        Ei := N2S(NEi);
        N2S(NVr,Vr);
        N2S(NVi,Vi);
      end;
    end if;
    Restore_Scalar_Mode;
    TraceLeave;
  end ApproxEigen;

  procedure ApproxEigen1(A: in SMatrix; E: in out Scalar; V: in out SVector) is
    Ei: Scalar := ScalZero;
    Vi: SVector(V'Range);
  begin
    E := ScalOne;
    ApproxEigen(A,E,Ei,V,Vi);
  end ApproxEigen1;

-----

  procedure DerMatrix(DY: in SDMap; P: in SParam; F: in SFourier2; A: in out SMatrix) is
  begin
    TraceEnter("Square.DerMatrix");
    Enter_Numeric_Mode;
    if Trunc then
      NumDerMatrix(DY,P,F,A);
    else
      declare
        NA: NMatrix(A'Range(1),A'Range(2));
      begin
        NContr.NumDerMatrix(S2N(DY),S2N(P),S2N(F),NA);
        N2S(NA,A);
      end;
    end if;
    Restore_Scalar_Mode;
    TraceLeave;
  end DerMatrix;

  procedure ContrMatrix(DY: in SDMap; P: in SParam; F: in SFourier2; A: in out SMatrix) is
  begin
    TraceEnter("Square.ContrMatrix");
    Enter_Numeric_Mode;
    if Trunc then
      NumContrMatrix(DY,P,F,A);
    else
      declare
        NA: NMatrix(A'Range(1),A'Range(2));
      begin
        NContr.NumContrMatrix(S2N(DY),S2N(P),S2N(F),NA);
        N2S(NA,A);
      end;
    end if;
    Restore_Scalar_Mode;
    TraceLeave;
  end ContrMatrix;

  procedure ApproxEigen1(DY: in SDMap; P: in SParam; F: in SFourier2; E: in out Scalar; U: in out SFourier2; Dim: in Positive) is
    V: SVector(1..Dim);
    A: SMatrix(1..Dim,1..Dim);
  begin
    DerMatrix(DY,P,F,A);
    ApproxEigen1(A,E,V);
    UnPack(V,F,U);
  end ApproxEigen1;

  procedure ApproxFix(Y: in SMap; DY: in SDMap; P: in SParam; F: in out SFourier2; A: in out SMatrix; Eps: in Radius) is
  begin
    TraceEnter("Square.ApproxFix");
    Enter_Numeric_Mode;
    if Trunc then
      NumFix(Y,DY,P,F,A,Eps);
    else
      declare
        NA: NMatrix(A'Range(1),A'Range(2));
        NF: NFourier2;
      begin
        S2N(F,NF);
        NContr.NumFix(S2N(Y),S2N(DY),S2N(P),NF,NA,Eps);
        N2S(NA,A);
        N2S(NF,F);
      end ;
    end if;
    Restore_Scalar_Mode;
    TraceLeave;
  end ApproxFix;

  function ContrFix(Y: SMap; DY: SDMap; P: SParam; F: SFourier2; DC: Positive; Eps: Radius) return SFourier2 is
    KMax: constant Radius := Flt(25)/Flt(32); -- can be changed
    A: SMatrix(1..DC,1..DC);
    G: SFourier2 := F;
  begin
    TraceEnter("Square.ContrFix");
    ApproxFix(Y,DY,P,G,A,Eps);
    if not Trunc then
      CheckFix(Y,DY,P,A,G,KMax);
      Show("ContrFix: verified solution");
    end if;
    TraceLeave;
    return G;
  end ContrFix;

  function ContrFixA(L: Scalar; F: SFourier2; DC: Positive; Eps: Radius) return SFourier2 is
    P: constant SParam2 := (L,ScalZero);
  begin
    return ContrFix(MapA'Access,DMapA'Access,P,F,DC,Eps);
  end ContrFixA;

  procedure CheckNoSymm(F: in SFourier2) is
    P1: constant Scalar := ScalQuarter*Pi;
    P2: constant Scalar := ScalTwo*P1;
    P3: constant Scalar := ScalThree*P1;
    V: SVector(1 .. 4);
  begin
    V(1) := abs(Val(P1,P2,F));
    V(2) := abs(Val(P2,P1,F));
    V(3) := abs(Val(P3,P2,F));
    V(4) := abs(Val(P2,P3,F));
    for I in 2 .. 4 loop
      for J in 1 .. I-1 loop
        if not (V(I) /= V(J)) then
          Error("CheckNoSymm error: could not verify asymmetry");
        end if;
      end loop;
    end loop;
    Show("CheckNoSymm: verified asymmetry");
  end CheckNoSymm;

------------------

  procedure BifFun(Y: in SMap; DY: in SDMap; L,M: in Scalar; U: in SFourier2; F: in out SFourier2; S: in out Scalar) is
    DC: constant Positive := 9;
    Eps: constant Radius := 1.0E-12;
    D: constant Positive := SomeDeg(U.C);
    V: ScalVec(-2..D);
    G: SFourier2;
  begin
    Pack(U,V);
    V( 0) := ScalOne;
    V(-1) := L;
    V(-2) := M;
    F := ContrFix(Y,DY,V,F,DC,Eps);
    V(0) := ScalZero;
    Y(V,F,G);
    S := Compo1(G,U)-M;
  end BifFun;

  procedure BifFunPlot(Y: in SMap; DY: in SDMap; L1,L2,M1,M2: in Flt; U,F0: in SFourier2; N: Positive := 10) is
    procedure LaMu(I,J: in Integer; L,M: out Flt) is
    begin
      L := (L1*Flt(N-I)+L2*Flt(N+I))/Flt(2*N);
      M := (M1*Flt(N-J)+M2*Flt(N+J))/Flt(2*N);
    end LaMu;
    L,M: Flt;
    S: Scalar;
    F1: SFourier2 := F0;
    F: SFourier2;
  begin
    for K in 0 .. N loop
      Show("preparation " & Strng(K));
      LaMu(-K,-K,L,M);
      BifFun(Y,DY,Scal(L),Scal(M),U,F1,S);
    end loop;
    Show("starting plot");
    for I in -N .. N loop
      LaMu(I,-N,L,M);
      BifFun(Y,DY,Scal(L),Scal(M),U,F1,S);
      Plot3(Standard_Output,L,M,Approx(S));
      F := F1;
      for J in 1-N .. N loop
        LaMu(I,J,L,M);
        BifFun(Y,DY,Scal(L),Scal(M),U,F,S);
        Plot3(Standard_Output,L,M,Approx(S));
      end loop;
      New_Line;
    end loop;
  end BifFunPlot;

------

  procedure BifFunA(L,M: in Scalar; U: in SFourier2; F: in out SFourier2; S: in out Scalar) is
  begin
    BifFun(BifMapA'Access,DBifMapA'Access,L,M,U,F,S);
  end BifFunA;

  procedure BifFunPlotA(L1,L2,M1,M2: in Flt; U,F0: in SFourier2; N: in Positive := 10) is
  begin
    BifFunPlot(BifMapA'Access,DBifMapA'Access,L1,L2,M1,M2,U,F0,N);
  end BifFunPlotA;

--------------------

  procedure CheckPositivity(D: in SDiag) is
  begin
    for I in D'Range loop
      if D(I)<ScalZero then
        Error("Square.CheckPositivity error");
      end if;
    end loop;
  end CheckPositivity;

  procedure DiagProd(W: in SVector; S: in Scalar; F: in SFourier2; G: in out SFourier2) is
--- diag(W) on low modes, S*Id on high modes
    V: SVector(W'Range);
  begin
    G := F;
    PackSplit(G,V);
    Mult(S,G);
    if not IsZero(V) then
      for K in V'Range loop Mult(W(K),V(K)); end loop;
      Add(UnPack(V,F),G);
    end if;
  end DiagProd;

  procedure IdProd(A: in SMatrix; S: in Scalar; F: in Fourier2; G: in out Fourier2) is
--- A on low modes, S*Id on high modes
     V: SVector(A'Range(2));
  begin
    G := F;
    PackSplit(G,V);
    Mult(S,G);
    if not IsZero(V) then
      Mult(A,V);
      Add(UnPack(V,F),G);
    end if;
  end IdProd;

  procedure DMapData(DY: in SDMap; P: in SParam; F: in SFourier2; T: in out SSimDiag) is
--- approx eigenvalues and eigenvectors of symmetrized DY
  begin
    TraceEnter("Square.DMapData");
    Enter_Numeric_Mode;
    if Trunc then
      NumEigenData(DY,P,F,T);
    else
      declare
        NT: NSimDiag(T.D'Last);
      begin
        NFouriers2.NumEigenData(S2N(DY),S2N(P),S2N(F),NT);
        N2S(NVector(NT.D),SVector(T.D));
        N2S(NT.L,T.L);
      end;
    end if;
    Restore_Scalar_Mode;
    Orthogonalize(T.L);
    Show("ortho err ",Sup(Norm1(OrthoErr(T.L))),1);
    TraceLeave;
  end DMapData;

  procedure DMapAData(L: in Scalar; F: in SFourier2; T: in out SSimDiag) is
    P: constant SParam2 := (L,ScalZero);
  begin
    DMapData(DMapA'Access,P,F,T);
    CheckPositivity(T.D);
  end DMapAData;

  function TT(D: SDiag; L: SMatrix) return SMatrix is
    A: SMatrix(D'Range,D'Range);
  begin
    LDLTransposed(D,L,A);
    UnSobolize(A);
    return A;
  end TT;

  function LowTTzInv(T: SSimDiag; Z: Scalar) return SMatrix is
    TD: SDiag renames T.D;
    DZ: SDiag(TD'Range);
  begin
    for I in TD'Range loop
      DZ(I) := Inv(TD(I)+Z);
    end loop;
    return TT(DZ,T.L);
  end LowTTzInv;

  procedure AA(L: in Scalar; T: in SMatrix; F: in SFourier2; H: in out SFourier2) is
--- DMapA-TT
    P: constant SParam2 := (L,ScalZero);
    G: SFourier2;
  begin
    Prod(T,H,G);
    DMapA(P,F,H);
    Sub(G,H);
  end AA;

  procedure NormPerturbA(L: in Scalar; F: in SFourier2; B: in SSimDiag) is
    N: constant Positive := B.D'Last;
    Small: constant Radius := Quarter;
    Exclude: constant Flt := One;
----
    function HighDeg(K: Positive) return Boolean is
    begin
      return (K>N) and then NewDeg(K);
    end HighDeg;
----
    K: Scalar;
    T: SMatrix := TT(B.D,B.L);
    Tzi: SMatrix := LowTTzInv(B,Scal(-Exclude));
----
    procedure SP(G: in Fourier2; H: in out Fourier2) is
    begin
      IdProd(Tzi,Inv(Scal(-Exclude)),G,H);
      AA(L,T,F,H);
    end SP;
----
    function SPNorm is new OperatorNorm(HighDeg => HighDeg, LowOp => SP, HighOp => SP);
  begin
    TraceEnter("Square.NormPerturbA");
    K := SPNorm(F,Small,One);
    if Sup(K)<One then
      Show("NormPerturbA: check passed");
    else
      Error("NormPerturbA error: K>=1");
    end if;
    TraceLeave;
  end NormPerturbA;

  procedure CheckEigen(L: in Scalar; F: in SFourier2; B: in SSimDiag) is
    D: SDiag renames B.D;
    Morse: Integer := 0;
  begin
    NormPerturbA(L,F,B);
    for I in D'Range loop
      if D(I)>ScalOne then
        Morse := Morse+1;
      end if;
    end loop;
    Show1("CheckEigen: Morse index is ",Morse);
  end CheckEigen;

begin

  Restore_Scalar_Mode;

end Square;
