with Ada.Text_IO, Ints, Flts, Symmetries;
use Ada.Text_IO, Ints, Flts, Symmetries;
with Constants, Vectors, Vectors.Ops, Vectors.Num, Fouriers2, NumFouriers2;

pragma Elaborate_All (NumFouriers2, Constants);

generic

  Deg: in Positive;

-------------------- begin_include reps NRep

  type NRep is private;  --- used to define intervals, balls, ...

  with function IsZero(R: NRep) return Boolean is <>;
  with procedure SetZero(R: in out NRep) is <>;
  with function Sign(R: NRep) return Integer is <>;
  with function "abs"(R: NRep) return NRep is <>;
  with function Min(R1,R2: NRep) return NRep is <>;
  with function Max(R1,R2: NRep) return NRep is <>;
  with procedure Swap(R1,R2: in out NRep) is <>;

  with function Compare(R1,R2: NRep) return Integer is <>;
  with function "<"(R1,R2: NRep) return Boolean is <>;
  with function "<="(R1,R2: NRep) return Boolean is <>;
  with function "="(R1,R2: NRep) return Boolean is <>;
  with function ">="(R1,R2: NRep) return Boolean is <>;
  with function ">"(R1,R2: NRep) return Boolean is <>;

  with function "-"(R: NRep) return NRep is <>;
  with function "+"(R1,R2: NRep) return NRep is <>;
  with function "-"(R1,R2: NRep) return NRep is <>;
  with function "*"(R1,R2: NRep) return NRep is <>;
  with function "/"(R1,R2: NRep) return NRep is <>;
  with function "**"(R: NRep; N: Integer) return NRep is <>;
  with function "*"(F: Flt; R: NRep) return NRep is <>;
  with function Sqrt(R: NRep) return NRep is <>;

  with function Upper(R: NRep) return Flt is <>;
  with function Lower(R: NRep) return Flt is <>;
  with function Approx(R: NRep) return Flt is <>;
  with function ApproxPut(R: NRep) return FltVector is <>;
  with function ApproxGet(V: FltVector) return NRep is <>;
  with function Val(I: Integer) return NRep is <>;
  with function Val(F: Flt) return NRep is <>;
  with function ValDec(S: String) return NRep is <>;
  with function ValHex(S: String) return NRep is <>;
  with function DecStr(R: NRep; Digs: Natural := 0) return String is <>; -- default, if NDig=0
  with function HexStr(R: NRep; Digs: Natural := 0) return String is <>; -- full, if NDig=0
  with function Rounded_Last_Conversion return Boolean is <>;

  with procedure Set_Precision(Dummy: in NRep; Binary_Digs: in Positive) is <>;
  with function Get_Precision(R: NRep) return Positive is <>;
  with procedure Set_Rounding_Up(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Down(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Nearest(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Truncate(Dummy: in NRep) is <>;
  with procedure Set_Rounding_Relaxed(Dummy: in NRep) is <>;

--- approximate
  with function Num_Exp(R: NRep) return NRep is <>;
  with function Num_Log(R: NRep) return NRep is <>;
  with function Num_Cos(R: NRep) return NRep is <>;
  with function Num_Sin(R: NRep) return NRep is <>;

-------------------- end_include reps NRep

-------------------- begin_include scalars Scalar

  type Scalar is private;

--- basic
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with function Copy(S: Scalar) return Scalar is <>;
  with procedure Enclosure(S: in out Scalar; Binary_Digs: in Positive) is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with function FPut(R: Scalar) return FltVector is <>;
  with function FGet(V: FltVector) return Scalar is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function Scal(R1,R2: Flt) return Scalar is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

--- intervals
  with function Sup(S: Scalar) return Flt is <>;
  with function Inf(S: Scalar) return Flt is <>;
  with function Disjoint(S1,S2: Scalar) return Boolean is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;

--- sets
  with function Ball0(S: Scalar) return Scalar is <>;
  with function Ball0(R: Flt) return Scalar is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ResetCenter(S: in out Scalar) is <>;
  with function Subset(S1,S2: Scalar) return Boolean is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with function Intersection(S1,S2: Scalar) return Scalar is <>;

--- ops
  with function Norm(S: Scalar) return Scalar is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Flt is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(S: in Scalar; T: in out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with procedure Sub(S: in Scalar; T: in out Scalar) is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Mult(S: in Scalar; T: in out Scalar) is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

--- fun
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Pi return Scalar is <>;

--- io
  with procedure ShowType(S: in Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;
  with procedure Write(Name: in String; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Read(Name: in String; S: in out Scalar; Decimal: in Boolean := False) is <>;

--- misc
  with procedure Set_Active_Mode(Dummy: in Scalar) is <>;
  with procedure Set_Wait_Mode(Dummy: in Scalar) is <>;
  with procedure Set_New_Precision(Dummy: in Scalar; Binary_Digs: in Positive) is <>;
  with function Get_Precision(S: Scalar) return Positive is <>;
-------------------- end_include scalars Scalar

package Square is

  pragma Elaborate_Body;

  package SConstants is new Constants (Scalar => Scalar);
  package SVectors is new Vectors (Scalar => Scalar);
  package SVectors_Ops is new SVectors.Ops;
  package SVectors_Num is new SVectors.Num;
  package SFouriers2 is new Fouriers2 (Deg1 => Deg, Deg2 => Deg, Scalar => Scalar,
          SC => SConstants, SV => SVectors, SV_Ops => SVectors_Ops, SV_Num => SVectors_Num);
  package SContr renames SFouriers2.Contr;

  subtype Freq1  is SFouriers2.Freq1;
  subtype Freq2  is SFouriers2.Freq2;
  subtype FreqE1 is SFouriers2.FreqE1;
  subtype FreqE2 is SFouriers2.FreqE2;
  subtype SFourier2 is SFouriers2.Fourier2;
  subtype SVector is SFouriers2.ScalVec;
  subtype SDiag is SFouriers2.ScalDiag;
  subtype SMatrix is SFouriers2.ScalMat;
  subtype SSimDiag is SFouriers2.ScalSimDiag;
  subtype SLU is SFouriers2.ScalLU;
  subtype SParam is SFouriers2.Param;
  subtype SParam2 is SParam(1..2);
  subtype SMap is SFouriers2.Map;
  subtype SDMap is SFouriers2.DMap;

  procedure Show_Param;
  procedure Set_Scalar_Precision(Binary_Digs: in Positive);
  procedure Set_Numeric_Precision(Binary_Digs: in Positive);
  procedure Set_Precision(Binary_Digs: in Positive);
  function Get_Scalar_Precision return Positive;
  function Get_Numeric_Precision return Positive;

  procedure ApproxEigen(A: in SMatrix; Wr,Wi: in out SVector; Vr,Vi: in out SMatrix);
  procedure ApproxEigen(A: in SMatrix; Er,Ei: in out Scalar; Vr,Vi: in out SVector);
  procedure ApproxEigen1(A: in SMatrix; E: in out Scalar; V: in out SVector);

  procedure DerMatrix(DY: in SDMap; P: in SParam; F: in SFourier2; A: in out SMatrix);
  procedure ContrMatrix(DY: in SDMap; P: in SParam; F: in SFourier2; A: in out SMatrix);
  procedure ApproxEigen1(DY: in SDMap; P: in SParam; F: in SFourier2; E: in out Scalar; U: in out SFourier2; Dim: in Positive);
  procedure ApproxFix(Y: in SMap; DY: in SDMap; P: in SParam; F: in out SFourier2; A: in out SMatrix; Eps: in Radius);
  function ContrFix(Y: SMap; DY: SDMap; P: SParam; F: SFourier2; DC: Positive; Eps: Radius) return SFourier2;
  function ContrFixA(L: Scalar; F: SFourier2; DC: Positive; Eps: Radius) return SFourier2;
  procedure CheckNoSymm(F: in SFourier2);

--- bifurcation related
  procedure BifFun(Y: in SMap; DY: in SDMap; L,M: in Scalar; U: in SFourier2; F: in out SFourier2; S: in out Scalar);
  procedure BifFunPlot(Y: in SMap; DY: in SDMap; L1,L2,M1,M2: in Flt; U,F0: in SFourier2; N: in Positive := 10);
  procedure BifFunA(L,M: in Scalar; U: in SFourier2; F: in out SFourier2; S: in out Scalar);
  procedure BifFunPlotA(L1,L2,M1,M2: in Flt; U,F0: in SFourier2; N: Positive := 10);

--- spectrum related stuff
  procedure CheckPositivity(D: in SDiag);
  procedure DiagProd(W: in SVector; S: in Scalar; F: in SFourier2; G: in out SFourier2);
  procedure IdProd(A: in SMatrix; S: in Scalar; F: in SFourier2; G: in out SFourier2);
  procedure DMapData(DY: in SDMap; P: in SParam; F: in SFourier2; T: in out SSimDiag);
  procedure DMapAData(L: in Scalar; F: in SFourier2; T: in out SSimDiag);
  function TT(D: SDiag; L: SMatrix) return SMatrix;
  function LowTTzInv(T: SSimDiag; Z: Scalar) return SMatrix;
  procedure AA(L: in Scalar; T: in SMatrix; F: in SFourier2; H: in out SFourier2);
  procedure NormPerturbA(L: in Scalar; F: in SFourier2; B: in SSimDiag);
  procedure CheckEigen(L: in Scalar; F: in SFourier2; B: in SSimDiag);

private

  package NF2 is new NumFouriers2 (Deg1 => Deg, Deg2 => Deg, NRep => NRep);
  package NFouriers2 renames NF2.NFouriers2;
  package NConstants renames NF2.NConstants;
  package NVectors renames NF2.NVectors;
  package NVectors_Ops renames NF2.NVectors_Ops;
  package NVectors_Num renames NF2.NVectors_Num;
  package NContr renames NFouriers2.Contr;
  use NF2, SFouriers2, SConstants, SVectors, SVectors_Ops, SContr;

  function R2S(R: FltVector) return Scalar renames FGet;
  function S2R(N: Scalar) return FltVector renames FPut;
  function N2S(N: Numeric) return Scalar;
  function S2N(S: Scalar) return Numeric;
  procedure N2S(V: in NVector; W: in out SVector);
  procedure S2N(V: in SVector; W: in out NVector);
  function S2N(V: SVector) return NVector;
  procedure N2S(A: in NMatrix; B: in out SMatrix);
  procedure S2N(A: in SMatrix; B: in out NMatrix);
  function S2N(A: SMatrix) return NMatrix;
  procedure N2S(F: in NFourier2; G: in out SFourier2);
  procedure S2N(F: in SFourier2; G: in out NFourier2);
  function S2N(F: SFourier2) return NFourier2;
  function S2N(Y: SMap) return NMap;
  function S2N(Y: SDMap) return NDMap;

  procedure Enter_Numeric_Mode;    -- both precision and rounding
  procedure Restore_Scalar_Mode;   -- both precision and rounding

  SDummy: Scalar renames SConstants.ScalZero;
  NDummy: Numeric renames NConstants.ScalZero;
  Scalar_Precision: Positive := Get_Precision(SDummy);
  Numeric_Precision: Positive := Numbers.Get_Precision(NDummy);

end Square;
