with Messages;
use Messages;

package body Symmetries is

  procedure Trivial(M,N: in Natural; Swap: out Boolean; Sig: out Flt) is
  begin
    Swap := False;
    Sig := One;
  end Trivial;

  procedure XReflect(M,N: in Natural; Swap: out Boolean; Sig: out Flt) is
--- only for parity (1,1)
  begin
    Swap := False;
    if (M mod 2)=0 then Sig := NegOne; else Sig := One; end if;
  end XReflect;

  procedure YReflect(M,N: in Natural; Swap: out Boolean; Sig: out Flt) is
--- only for parity (1,1)
  begin
    Swap := False;
    if (N mod 2)=0 then Sig := NegOne; else Sig := One; end if;
  end YReflect;

  procedure DiagReflect(M,N: in Natural; Swap: out Boolean; Sig: out Flt) is
  begin
    Swap := (M /= N);
    Sig := One;
  end DiagReflect;

  procedure RotatePi(M,N: in Natural; Swap: out Boolean; Sig: out Flt) is
  begin
    Swap := False;
    if (M mod 2)=(N mod 2) then Sig := NegOne; else Sig := One; end if;
  end RotatePi;

  procedure NextMode(M,N: in Integer; K: in out Integer) is
  begin
    if PackSymmetry /= Reflect then
      MakePackIndex;
    end if;
    K := K+1;
    if K <= NPack then
      PackIndex(K).M := M;
      PackIndex(K).N := N;
      PackIndex(K).D := M+N;
    end if;
  end NextMode;

  procedure MakePackIndex is
    Swap: Boolean;
    K: Integer := 0;
    N: Natural;
    Sig: Flt;
  begin
    PackSymmetry := Reflect;
    for D in 2 .. Integer'Last loop
      for M in 1 .. D-1 loop
        N := D-M;
        Reflect(M,N,Swap,Sig);
        if not Swap then
          if Sig>Zero then NextMode(M,N,K); end if;
        elsif M<N then
          NextMode(M,N,K);
        end if;
        if K=NPack then return; end if;
      end loop;
    end loop;
  end MakePackIndex;

  procedure Coord(K: in Positive; M,N: out Natural) is
    P: ModeInfo renames PackIndex(K);
  begin
    if PackSymmetry /= Reflect then
      MakePackIndex;
    end if;
    M := P.M;
    N := P.N;
  end Coord;

  function Deg(K: Positive) return Positive is
  begin
    if PackSymmetry /= Reflect then
      MakePackIndex;
    end if;
    return PackIndex(K).D;
  end Deg;

  function MaxInvDeg(D: Positive) return Positive is
    K: Integer := 1;
  begin
    if PackSymmetry /= Reflect then
      MakePackIndex;
    end if;
    while PackIndex(K).D <= D loop K := 2*K; end loop;
    K := K/2;
    while PackIndex(K).D <= D loop K := K+1; end loop;
    return K-1;
  end MaxInvDeg;

  procedure Show(M,N: in Natural; Sig: in Flt; NewLine: in Boolean) is
  begin
    Show0("  [" & Strng(M,4) & "," & Strng(N,4) & ",",False);
    if Sig>Zero then
      Show0(" + ]",NewLine);
    else
      Show0(" - ]",NewLine);
    end if;
  end Show;

  procedure ShowModes(NMax: in Positive) is
    Swap: Boolean;
    M,N: Natural;
    Sig: Flt;
  begin
    for K in 1 .. NMax loop
      Coord(K,M,N);
      Reflect(M,N,Swap,Sig);
      Show0(Strng(K,4),False);
      if Swap then
        Show(M,N,One,False);
      end if;
      Show(M,N,Sig,True);
    end loop;
  end ShowModes;

end Symmetries;
