with Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO, Symmetries, Numerics, Square;
use Ada.Text_IO, Messages, Ints, Ints.IO, Flts, Flts.IO, Symmetries;

--with Long_Floats; use Long_Floats;
with MPFR.Floats; use MPFR.Floats;

procedure TestingA is

--  type NRep is new Rep;  -- rep used for numeric computations
--  type SRep is new Rep;  -- rep used to define Scalar

  type NRep is new MPFloat;  -- rep used for numeric computations
  type SRep is new MPFloat;  -- rep used to define Scalar

  package Scalars is new Numerics (Rep => SRep);
  use Scalars;
  type Scalar is new Numeric;

  Deg: constant Positive := GetArg;

  package SQ is new Square (Deg => Deg, NRep => NRep, Scalar => Scalar);
  use SQ;
  use SConstants, SVectors, SFouriers2, SContr;
  subtype Fun is SFourier2;

  Nop: constant Symmetry := Symmetries.Trivial'Access;
  Ref: Symmetry renames Symmetries.Reflect;

  DC:     constant Integer := 9;           -- used in contraction
  LSize:  constant Integer := 150;         -- used for eigenvalues/vectors
  Eps:    constant Radius  := 1.0E-20;     -- for approx solutions
  Rad:    constant Radius  := 1.0/8.0;     -- domain
--  Lambda: constant Flt     := GetArg;
  Lambda: constant Flt     := 145.0/256.0;
  P:      constant SParam2 := (Scal(Lambda),ScalZero);

  procedure ShowParam is
  begin
    Show_Param;
    Show1("    DC =",DC);
    Show1(" LSize =",LSize);
    Show1("   Eps =",Eps);
    Show1("Lambda =",Lambda);
    Show1("   Rad =",Rad);
  end ShowParam;

  procedure ShowEigen(N: in Positive; F: in Fun) is
    RefSave: constant Symmetry := Ref;
    Wr,Wi: SVector(1..LSize);
    A,Vr,Vi: SMatrix(1..LSize,1..LSize);
  begin
    Ref := Nop;                     --- no symmetry
    DerMatrix(DMapA'Access,P,F,A);
    ApproxEigen(A,Wr,Wi,Vr,Vi);
    for J in 1 .. N loop
      Show1("  E" & Strng(J,1),Approx(Wr(J)));
--      Write("eigen" & Strng(J,1) & ".data",UnPack(GetColumn(J,Vr),F));
    end loop;
    Ref := RefSave;
  end ShowEigen;

  procedure FindFix(Bif: in Flt) is
    R0: constant RadPair := (Rad,Rad);
    E: Scalar;
    F0,F1,U: Fun;
  begin
    Read("fix-in.data",F0);
    F0.T.R := R0;
    Write("perturb.data",Perturb(R0,P(1)));
    if Bif /= Zero then
      ApproxEigen1(DMapA'Access,P,F0,E,U,LSize);
      Write("eigen.data",U);
      F0 := F0+Scal(Bif)*U;                -- used for bifurcation
    end if;
    F1 := ContrFixA(P(1),F0,DC,Eps);
    Write("fix-out.data",F1);
    Show1("norm",Norm(F1));
    MapA(P,F1,F0);
    Show1(" err",Norm(F0-F1));
    ShowEigen(5,F1);
  end FindFix;

begin
  Verbosity := 1;
  Set_Precision(128);
  ShowParam;

  Ref := Symmetries.Trivial'Access;     -- no symmetry
--  Ref := Symmetries.DiagReflect'Access; -- impose symmetry

--  FindFix(0.1);  -- used to follow new branches
  FindFix(0.0);  -- used to improve approx solutions

end TestingA;
