with Ada.Text_IO, Messages;
use Ada.Text_IO, Messages;

package body Vectors.Num is

  ScalZero:  constant Scalar := Scal(Zero);
  ScalOne:   constant Scalar := Scal(One);
  ScalTwo:   constant Scalar := Scal(Two);
  ScalHalf:  constant Scalar := Scal(Half);

  function CountLarge(B: Vector; Cut: Scalar) return Integer is
    N: Integer := 0;
  begin
    for I in B'Range loop
      if B(I)>Cut then N := N+1; end if;
    end loop;
    return N;
  end CountLarge;

--- linpack inverse* for real matrices ----------------------------------------------

   procedure dgefa(n: in Integer; a: in out Matrix; ipvt: out Pivot; info: out Integer) is
-- translated from linpack routine dgefa, http://www.netlib.org/linpack/
      l: Integer;
      t: Scalar;

   begin
      info := 0;
      for k in 1 .. n-1 loop

--        find l := pivot index;

         t := abs(a(n,k));
         l := n;
         for m in reverse k .. n-1 loop
            if (abs(a(m,k)) >= t) then
               t := abs(a(m,k));
               l := m;
            end if;
         end loop;
         ipvt(k) := l;

--        zero pivot implies this column already triangularized

         if IsZero(a(l,k)) then
            info := k;
         else

--           interchange if necessary

            if (l /= k) then
               Swap(a(l,k),a(k,k));
            end if;

--           compute multipliers

            t := Scal(-1)/a(k,k);
            for m in k+1 .. n loop
               a(m,k) := t*a(m,k);
            end loop;

--           row elimination with column indexing

            for j in k+1 .. n loop
               t := a(l,j);
               if (l /= k) then
                  Swap(a(l,j),a(k,j));
               end if;
               for m in k+1 .. n loop
                  a(m,j) := a(m,j) + t*a(m,k);
               end loop;
            end loop;
         end if;
      end loop;

      ipvt(n) := n;
      if IsZero(a(n,n)) then info := n; end if;
   end dgefa;

   procedure dgesl(n,job: in Integer; ipvt: in Pivot; a: in Matrix; b: in out Vector) is
-- translated from linpack routine dgesl, http://www.netlib.org/linpack/
      l: Integer;
      t: Scalar;

   begin
      if (job = 0) then

--        job = 0 , solve  a * x := b;
--        first solve  l*y := b;

         for k in 1 .. n-1 loop
            l := ipvt(k);
            t := b(l);
            if (l /= k) then
               Swap(b(l),b(k));
            end if;
            for m in k+1 .. n loop
               b(m) := b(m) + t*a(m,k);
            end loop;
         end loop;

--        now solve  u*x := y;

         for k in reverse 1 .. n loop
            b(k) := b(k)/a(k,k);
            t := -b(k);
            for m in 1 .. k-1 loop
               b(m) := b(m) + t*a(m,k);
            end loop;
         end loop;
         return;
      end if;

--     job = nonzero, solve  trans(a) * x := b;
--     first solve  trans(u)*y := b;

      for k in 1 .. n loop
         t := ScalZero;
         for m in 1 .. k-1 loop
            t := t + a(m,k)*b(m);
         end loop;
         b(k) := (b(k) - t)/a(k,k);
      end loop;

--     now solve trans(l)*x := y;

      for k in reverse 1 .. n-1 loop
         t := b(k);
         for m in k+1 .. n loop
            t := t + a(m,k)*b(m);
         end loop;
         b(k) := t;
         l := ipvt(k);
         if (l /= k) then
            Swap(b(l),b(k));
         end if;
      end loop;
   end dgesl;

   procedure dgedi(n,job: in Integer; ipvt: in Pivot; a: in out Matrix; work: in out Vector; det: in out Scalar) is
-- translated from linpack routine dgedi, http://www.netlib.org/linpack/
      l: integer;
      t: Scalar;

   begin
--     compute determinant

      if (job /= 1) then
         det := ScalOne;
         for i in 1 .. n loop
            if (ipvt(i) /= i) then det := -det; end if;
            det := a(i,i)*det;
            exit when IsZero(det);
         end loop;
      end if;

--     compute inverse(u)

      if (job = 10) then return; end if;

      for k in 1 .. n loop
         a(k,k) := ScalOne/a(k,k);
         t := -a(k,k);
         for m in 1 .. k-1 loop
            a(m,k) := t*a(m,k);
         end loop;
         for j in k+1 .. n loop
            t := a(k,j);
            a(k,j) := ScalZero;
            for m in 1 .. k loop
               a(m,j) := a(m,j) + t*a(m,k);
            end loop;
         end loop;
      end loop;

--     form inverse(u)*inverse(l)

      for k in reverse 1 .. n-1 loop
         for i in k+1 .. n loop
            work(i) := a(i,k);
            a(i,k) := ScalZero;
         end loop;
         for j in k+1 .. n loop
            t := work(j);
            for m in 1 .. n loop
               a(m,k) := a(m,k) + t*a(m,j);
            end loop;
         end loop;
         l := ipvt(k);
         if (l /= k) then
            for m in 1 .. n loop
               Swap(a(m,k),a(m,l));
            end loop;
         end if;
      end loop;
   end dgedi;

  procedure Solve(A: in Matrix; X: in out Vector) is
    pragma Suppress(Storage_Check);
    N: constant Integer := A'Last(1);
    Info: Integer;
    Ipvt: Pivot(1 .. N);
    Ar: Matrix(1 .. N,1 .. N);
  begin
    Ar := A;
    Dgefa(N,Ar,Ipvt,Info);
    if Info>0 then
      Show("Warning: Problem with diagonal ",Info);
    end if;
    Dgesl(N,0,Ipvt,Ar,X);
  end Solve;

  procedure TSolve(A: in Matrix; X: in out Vector) is
    pragma Suppress(Storage_Check);
    N: constant Integer := A'Last(1);
    Info: Integer;
    Ipvt: Pivot(1 .. N);
    Ar: Matrix(1 .. N,1 .. N);
  begin
    Ar := A;
    Dgefa(N,Ar,Ipvt,Info);
    if Info>0 then
      Show("Warning: Problem with diagonal ",Info);
    end if;
    Dgesl(N,1,Ipvt,Ar,X);
  end TSolve;

  procedure Invert(A: in out Matrix) is
    pragma Suppress(Storage_Check);
    N: constant Integer := A'Last(1);
    Info: Integer;
    Det: Scalar;
    Ipvt: Pivot(1 .. N);
    Work: Vector(1 .. N);
  begin
    Dgefa(N,A,Ipvt,Info);
    if Info>0 then
      Show("Warning: Problem with diagonal ",Info);
    end if;
    Dgedi(N,1,Ipvt,A,Work,Det);
  end Invert;

  function Inverse(A: Matrix) return Matrix is
    B: Matrix := A;
  begin
    Invert(B);
    return B;
  end Inverse;

--- eispack eigen* for real symmetric matrices --------------------------------------

   function dsign(a,b: Scalar) return Scalar is
---  ancient fortran routine
   begin
      if B<ScalZero then return -abs(A); end if;
      return abs(A);
   end dsign;

   procedure tred1(n: in Integer; a: in out Matrix; d,e,e2: in out Vector) is
-- translated from eispack routine tred1, http://www.netlib.org/eispack/
      l: Integer;
      f,g,h,scale: Scalar;

begin

      for i in 1 .. n loop
         d(i) := a(n,i);
         a(n,i) := a(i,i);
      end loop;

      for i in reverse 1 .. n loop
         l := i - 1;
         h := ScalZero;
         scale := ScalZero;

         if (l < 1) then
            e(i) := ScalZero;
            e2(i) := ScalZero;
            goto L300;
         else

--     .......... scale row (algol tol then not needed) ..........

            for k in 1 .. l loop
               scale := scale + abs(d(k));
            end loop;
            if (scale = ScalZero) then
               for j in 1 .. l loop
                  d(j) := a(l,j);
                  a(l,j) := a(i,j);
                  a(i,j) := ScalZero;
               end loop;
               e(i) := ScalZero;
               e2(i) := ScalZero;
               goto L300;
            end if;
         end if;

         for k in 1 .. l loop
            d(k) := d(k) / scale;
            h := h + d(k) * d(k);
         end loop;
--
         e2(i) := scale * scale * h;
         f := d(l);
         g := -dsign(sqrt(h),f);
         e(i) := scale * g;
         h := h - f * g;
         d(l) := f - g;
         if (l /= 1) then

--     .......... form a*u ..........

            for j in 1 .. l loop
               e(j) := ScalZero;
            end loop;

            for j in 1 .. l loop
               f := d(j);
               g := e(j) + a(j,j) * f;
               if (l >= j+1) then
                  for k in j+1 .. l loop
                     g := g + a(k,j) * d(k);
                     e(k) := e(k) + a(k,j) * f;
                  end loop;
               end if;
               e(j) := g;
            end loop;

--     .......... form p ..........

            f := ScalZero;
            for j in 1 .. l loop
               e(j) := e(j) / h;
               f := f + e(j) * d(j);
            end loop;
            h := f / (h + h);

--     .......... form q ..........

            for j in 1 .. l loop
               e(j) := e(j) - h * d(j);
            end loop;

--     .......... form reduced a ..........

            for j in 1 .. l loop
               f := d(j);
               g := e(j);
               for k in j .. l loop
                  a(k,j) := a(k,j) - f * e(k) - g * d(k);
               end loop;
            end loop;
         end if;

         for j in 1 .. l loop
            f := d(j);
            d(j) := a(l,j);
            a(l,j) := a(i,j);
            a(i,j) := f * scale;
         end loop;

  <<L300>> null;

      end loop;
   end tred1;

   procedure tred2(n: in Integer; a: in Matrix; d,e: in out Vector; z: in out Matrix) is
-- translated from eispack routine tql1, http://www.netlib.org/eispack/
      l: integer;
      f,g,h,hh,scale: Scalar;

   begin
      if (n = 1) then
         d(1) := a(1,1);
         z(1,1) := ScalOne;
         e(1) := ScalZero;
         return;
      end if;

      for i in 1 .. n loop
         for j in i .. n loop
            z(j,i) := a(j,i);
         end loop;
         d(i) := a(n,i);
      end loop;

      for i in reverse 2 .. n loop
         l := i - 1;
         h := ScalZero;
         scale := ScalZero;

         if (l < 2) then
            e(i) := d(l);
            for j in 1 .. l loop
               d(j) := z(l,j);
               z(i,j) := ScalZero;
               z(j,i) := ScalZero;
            end loop;
            goto L290;
         else

--     .......... scale row (algol tol then not needed) ..........

            for k in 1 .. l loop
               scale := scale + abs(d(k));
            end loop;
            if (scale = ScalZero) then
               e(i) := d(l);
               for j in 1 .. l loop
                  d(j) := z(l,j);
                  z(i,j) := ScalZero;
                  z(j,i) := ScalZero;
               end loop;
               goto L290;
            end if;
         end if;

         for k in 1 .. l loop
            d(k) := d(k) / scale;
            h := h + d(k) * d(k);
         end loop;

         f := d(l);
         g := -dsign(sqrt(h),f);
         e(i) := scale * g;
         h := h - f * g;
         d(l) := f - g;

--     .......... form a*u ..........

         for j in 1 .. l loop
            e(j) := ScalZero;
         end loop;

         for j in 1 .. l loop
            f := d(j);
            z(j,i) := f;
            g := e(j) + z(j,j) * f;
            if (l > j) then
               for k in j+1 .. l loop
                  g := g + z(k,j) * d(k);
                  e(k) := e(k) + z(k,j) * f;
               end loop;
            end if;
            e(j) := g;
         end loop;

--     .......... form p ..........

         f := ScalZero;

         for j in 1 .. l loop
            e(j) := e(j) / h;
            f := f + e(j) * d(j);
         end loop;
         hh := f / (h + h);

--     .......... form q ..........

         for j in 1 .. l loop
            e(j) := e(j) - hh * d(j);
         end loop;

--     .......... form reduced a ..........

         for j in 1 .. l loop
            f := d(j);
            g := e(j);
            for k in j .. l loop
               z(k,j) := z(k,j) - f * e(k) - g * d(k);
            end loop;
            d(j) := z(l,j);
            z(i,j) := ScalZero;
         end loop;

  <<L290>> null;

         d(i) := h;
      end loop;

--     .......... accumulation of transformation matrices ..........

      for i in 2 .. n loop
         l := i - 1;
         z(n,l) := z(l,l);
         z(l,l) := ScalOne;
         h := d(i);
          if (h /= ScalZero) then
            for k in 1 .. l loop
               d(k) := z(k,i) / h;
            end loop;

            for j in 1 .. l loop
               g := ScalZero;
               for k in 1 .. l loop
                  g := g + z(k,i) * z(k,j);
               end loop;
               for k in 1 .. l loop
                  z(k,j) := z(k,j) - g * d(k);
               end loop;
            end loop;
         end if;
         for k in 1 .. l loop
            z(k,i) := ScalZero;
         end loop;
      end loop;

      for i in 1 .. n loop
         d(i) := z(n,i);
         z(n,i) := ScalZero;
      end loop;
      z(n,n) := ScalOne;
      e(1) := ScalZero;
   end tred2;

   procedure tql1(n: in Integer; d,e: in out Vector; ierr: out Integer) is
-- translated from eispack routine tql1, http://www.netlib.org/eispack/
     j,m,l1,l2: integer;
     c,c2,c3,dl1,el1,f,g,h,p,r,s,s2,tst1,tst2: Scalar;

   begin
      ierr := 0;
      if (n = 1) then return; end if;
      for i in 2 .. n loop
          e(i-1) := e(i);
      end loop;
      f := ScalZero;
      tst1 := ScalZero;
      e(n) := ScalZero;

      for l in 1 .. n loop
         j := 0;
         h := abs(d(l)) + abs(e(l));
         if (tst1 < h) then tst1 := h; end if;

--     .......... look for small sub-diagonal element ..........

         m := l;
         loop
            tst2 := tst1 + abs(e(m));
            exit when (tst2 = tst1);
            m := m+1;
         end loop;

         if (m /= l) then
            loop
               if (j = 30) then

--     .......... set error -- no convergence to an
--                eigenvalue after 30 iterations ..........

                  ierr := l;
                  return;
               end if;

               j := j + 1;

--     .......... form shift ..........

               l1 := l + 1;
               l2 := l1 + 1;
               g := d(l);
               p := (d(l1) - g) / (ScalTwo * e(l));
               r := sqrt(ScalOne+p*p);
               d(l) := e(l) / (p + dsign(r,p));
               d(l1) := e(l) * (p + dsign(r,p));
               dl1 := d(l1);
               h := g - d(l);
               for i in l2 .. n loop
                  d(i) := d(i) - h;
               end loop;
               f := f + h;

--     .......... ql transformation ..........

               p := d(m);
               c := ScalOne;
               c2 := c;
               el1 := e(l1);
               s := ScalZero;
               for i in reverse l .. m-1 loop
                  c3 := c2;
                  c2 := c;
                  s2 := s;
                  g := c * e(i);
                  h := c * p;
                  r := sqrt(p*p+e(i)*e(i));
                  e(i+1) := s * r;
                  s := e(i) / r;
                  c := p / r;
                  p := c * d(i) - s * g;
                  d(i+1) := h + s * (c * g + s * d(i));
               end loop;
--
               p := -s * s2 * c3 * el1 * e(l) / dl1;
               e(l) := s * p;
               d(l) := c * p;
               tst2 := tst1 + abs(e(l));
               exit when (tst2 = tst1);
            end loop;
         end if;

         p := d(l) + f;

--     .......... order eigenvalues ..........

         if (l = 1) then d(1) := p; end if;

         for i in reverse 2 .. l loop
            if (p >= d(i-1)) then
               d(i) := p;
               exit;
             end if;
            d(i) := d(i-1);
            if (i = 2) then d(1) := p; end if;
         end loop;
      end loop;
   end tql1;

   procedure tql2(n: in Integer; d,e: in out Vector; z: in out Matrix; ierr: out Integer) is
-- translated from eispack routine tql2, http://www.netlib.org/eispack/
      j,m,l1,l2: Integer;
      c,c2,c3,dl1,el1,f,g,h,p,r,s,s2,tst1,tst2: Scalar;

   begin
      ierr := 0;
      if (n = 1) then return; end if;
      for i in 2 .. n loop
         e(i-1) := e(i);
      end loop;
      f := ScalZero;
      tst1 := ScalZero;
      e(n) := ScalZero;

      for l in 1 .. n loop
         j := 0;
         h := abs(d(l)) + abs(e(l));
         if (tst1 < h) then tst1 := h; end if;

--     .......... look for small sub-diagonal element ..........

         m := l;
         loop
            tst2 := tst1 + abs(e(m));
            exit when (tst2 = tst1);
            m := m+1;
         end loop;

         if (m /= l) then
            loop
               if (j = 30) then

--     .......... set error -- no convergence to an
--                eigenvalue after 30 iterations ..........

                  ierr := l;
                  return;
               end if;

               j := j + 1;

--     .......... form shift ..........

               l1 := l + 1;
               l2 := l1 + 1;
               g := d(l);
               p := (d(l1) - g) / (ScalTwo * e(l));
               r := sqrt(ScalOne+p*p);
               d(l) := e(l) / (p + dsign(r,p));
               d(l1) := e(l) * (p + dsign(r,p));
               dl1 := d(l1);
               h := g - d(l);
               for i in l2 .. n loop
                  d(i) := d(i) - h;
               end loop;
               f := f + h;

--     .......... ql transformation ..........

               p := d(m);
               c := ScalOne;
               c2 := c;
               el1 := e(l1);
               s := ScalZero;
               for i in reverse l .. m-1 loop
                  c3 := c2;
                  c2 := c;
                  s2 := s;
                  g := c * e(i);
                  h := c * p;
                  r := sqrt(p*p+e(i)*e(i));
                  e(i+1) := s * r;
                  s := e(i) / r;
                  c := p / r;
                  p := c * d(i) - s * g;
                  d(i+1) := h + s * (c * g + s * d(i));

--     .......... form vector ..........
                  for k in 1 .. n loop
                     h := z(k,i+1);
                     z(k,i+1) := s * z(k,i) + c * h;
                     z(k,i) := c * z(k,i) - s * h;
                  end loop;
               end loop;

               p := -s * s2 * c3 * el1 * e(l) / dl1;
               e(l) := s * p;
               d(l) := c * p;
               tst2 := tst1 + abs(e(l));
               exit when (tst2 = tst1);
            end loop;
         end if;

         d(l) := d(l) + f;
      end loop;

--     .......... order eigenvalues and eigenvectors ..........

      for i in 1 .. n-1 loop
         m := i;
         p := d(i);
         for l in i+1 .. n loop
            if (d(l) < p) then
               m := l;
               p := d(l);
            end if;
         end loop;
         if (m /= i) then
            d(m) := d(i);
            d(i) := p;
            for l in 1 .. n loop
               p := z(l,i);
               z(l,i) := z(l,m);
               z(l,m) := p;
            end loop;
         end if;
      end loop;
   end tql2;

   procedure SortEigen(W: in out Vector; V: in out Matrix) is
     X,Y: Integer;
     A: Pivot(W'Range);
     U: Vector(W'Range);

     function GT(I,J: Integer) return Boolean is
     begin
       return abs(W(I))>abs(W(J));
     end GT;

     procedure Qsort(L,R: Integer) is
       I,J: integer;
     begin
       I:=L; J:=R; X:=A((L+R)/2);
       loop
         while GT(A(I),X) loop I:=I+1; end loop;
         while GT(X,A(J)) loop J:=J-1; end loop;
         if I <= J then
           Y := A(I); A(I) := A(J); A(J) := Y;
           I:=I+1; J:=J-1;
         end if;
         exit when I>J;
       end loop;
       if L<J then Qsort(L,J); end if;
       if I<R then Qsort(I,R); end if;
     end Qsort;

   begin
     for J in W'Range loop A(J) := J; end loop;
     Qsort(W'First,W'Last);

     U := W;
     for J in W'Range loop W(J) := U(A(J)); end loop;

     for I in W'Range loop
       for J in W'Range loop U(J) := V(I,J); end loop;
       for J in W'Range loop V(I,J) := U(A(J)); end loop;
     end loop;
   end SortEigen;

   procedure Eigen(A: in Matrix; W: in out Vector) is
     pragma Suppress(Storage_Check);
     N: constant Integer := A'Last(1);
     Ierr: Integer;
     Fv1,Fv2: Vector(1 .. N);
     Ar: Matrix(1 .. N,1 .. N);
   begin
     Ar := A;
     Tred1(N,Ar,W,Fv1,Fv2);
     Tql1(N,W,Fv1,Ierr);
     if Ierr>0 then
       Show("Error: Problem with eigenvalue number ",Ierr);
     end if;
   end Eigen;

   procedure Eigen(A: in Matrix; W: in out Vector; V: in out Matrix) is
     pragma Suppress(Storage_Check);
     N: constant Integer := A'Last(1);
     Ierr: Integer;
     Fv1: Vector(1 .. N);
     Ar: Matrix(1 .. N,1 .. N);
   begin
     Ar := A;
     Tred2(N,Ar,W,Fv1,V);
     Tql2(N,W,Fv1,V,Ierr);
     if Ierr>0 then
       Show("Error: Problem with eigenvalue number ",Ierr);
     end if;
     SortEigen(W,V);
   end Eigen;

-- eispack eigen* for complex matrices --------------------------------------------

   procedure cdiv(ar,ai,br,bi: in Scalar; cr,ci: in out Scalar) is
-- complex division, eispack style
      s: constant Scalar := br*br+bi*bi;
   begin
      cr := (ar*br+ai*bi)/s;
      ci := (ai*br-ar*bi)/s;
   end cdiv;

   procedure csroot(xr,xi: in Scalar; yr,yi: in out Scalar) is
-- complex square root, eispack style
-- branch chosen so that yr>=0 and sign(yi)=sign(xi)
      s: Scalar;
   begin
      s := sqrt(ScalHalf*(sqrt(xr*xr+xi*xi)+abs(xr)));
      if xr >= ScalZero then yr := s;  end if;
      if xi <  ScalZero then s  := -s; end if;
      if xr <= ScalZero then yi := s;  end if;
      if xr <  ScalZero then yr := ScalHalf*(xi/yi); end if;
      if xr >  ScalZero then yi := ScalHalf*(xi/yr); end if;
   end csroot;

   procedure cbal(n: in Integer; ar,ai: in out Matrix; low,igh: out Integer; scale: in out Vector) is
-- translated from eispack routine cbal, http://www.netlib.org/eispack/
     allzero,noconv: Boolean;
     k,l,m: Integer;
     c,f,g,r,s,b2,radix: Scalar;

--     .......... row and column exchange ..........
      procedure exc(m,j: in integer) is
      begin
         scale(m) := Scal(j);
         if (j = m) then return; end if;
         for i in 1 .. l loop
            Swap(ar(i,j),ar(i,m));
            Swap(ai(i,j),ai(i,m));
         end loop;
         for i in k .. n loop
            Swap(ar(j,i),ar(m,i));
            Swap(ai(j,i),ai(m,i));
         end loop;
      end exc;

   begin
      radix := ScalTwo;
      b2 := radix * radix;
      k := 1;
      l := n;

--     .......... search for rows isolating an eigenvalue
--                and push them down ..........
  <<L100>> null;

      for j in reverse 1 .. l loop
         allzero := True;
         for i in 1 .. l loop
            if (i /= j) then
               allzero := IsZero(ar(j,i)) and then IsZero(ai(j,i));
               exit when not allzero;
            end if;
         end loop;
         if allzero then
            m := l;
            exc(m,j);
            if (l = 1) then
               low := k;
               igh := l;
               return;
            end if;
            l := l - 1;
            goto L100;
         end if;
      end loop;

--     .......... search for columns isolating an eigenvalue
--                and push them left ..........
  <<L140>> null;

      for j in k .. l loop
         allzero := True;
         for i in k .. L loop
            if (i /= j) then
               allzero := IsZero(ar(i,j)) and then IsZero(ai(i,j));
               exit when not allzero;
            end if;
         end loop;
         if allzero then
            m := k;
            exc(m,j);
            k := k + 1;
            goto L140;
         end if;
      end loop;

--     .......... now balance the submatrix in rows k to l ..........
      for i in k .. l loop
         scale(i) := ScalOne;
      end loop;

--     .......... iterative loop for norm reduction ..........
  <<L190>> noconv := false;

      for i in k .. l loop
         c := ScalZero;
         r := ScalZero;
--
         for j in k .. l loop
            if (j /= i) then
               c := c + abs(ar(j,i)) + abs(ai(j,i));
               r := r + abs(ar(i,j)) + abs(ai(i,j));
            end if;
         end loop;

--     .......... guard against zero c or r due to underflow ..........
         if (c /= ScalZero) and (r /= ScalZero) then
            g := r / radix;
            f := ScalOne;
            s := c + r;
            while (c < g) loop
               f := f * radix;
               c := c * b2;
            end loop;
            g := r * radix;
            while (c >= g) loop
               f := f / radix;
               c := c / b2;
            end loop;

--     .......... now balance ..........
            if ((c + r) / f < Scal(0.95) * s) then
               g := ScalOne / f;
               scale(i) := scale(i) * f;
               noconv := true;
               for j in k .. n loop
                  ar(i,j) := ar(i,j) * g;
                  ai(i,j) := ai(i,j) * g;
               end loop;
               for j in 1 .. l loop
                  ar(j,i) := ar(j,i) * f;
                  ai(j,i) := ai(j,i) * f;
               end loop;
            end if;
         end if;
      end loop;

      if noconv then goto L190; end if;
      low := k;
      igh := l;
   end cbal;

   procedure cbabk2(n,low,igh,m: in Integer; scale: in Vector; zr,zi: in out Matrix) is
     k: Integer;
     s: Scalar;

   begin
      if (m = 0) then return; end if;

      if (igh /= low) then
         for i in low .. igh loop
            s := scale(i);
            for j in 1 .. m loop
               zr(i,j) := zr(i,j) * s;
               zi(i,j) := zi(i,j) * s;
            end loop;
         end loop;
      end if;

      for i in reverse 1 .. low-1 loop
         k := integer(Approx(scale(i)));
         if (k /= i) then
            for j in 1 .. m loop
               Swap(zr(i,j),zr(k,j));
               Swap(zi(i,j),zi(k,j));
            end loop;
         end if;
      end loop;

      for i in igh+1 .. n loop
         k := integer(Approx(scale(i)));
         if (k /= i) then
            for j in 1 .. m loop
               Swap(zr(i,j),zr(k,j));
               Swap(zi(i,j),zi(k,j));
            end loop;
         end if;
      end loop;
   end cbabk2;

   procedure comhes(n,low,igh: in Integer; ar,ai: in out Matrix; int: out Pivot) is
-- translated from eispack routine comhes, http://www.netlib.org/eispack/
      i,la,kp1,mm1,mp1: Integer;
      xr,xi,yr,yi: Scalar;

   begin
      la := igh - 1;
      kp1 := low + 1;
--
      for m in kp1 .. la loop
         mm1 := m - 1;
         xr := ScalZero;
         xi := ScalZero;
         i := m;
--
         for j in m .. igh loop
            if (abs(ar(j,mm1)) + abs(ai(j,mm1)) > abs(xr) + abs(xi)) then
               xr := ar(j,mm1);
               xi := ai(j,mm1);
               i := j;
            end if;
         end loop;
--
         int(m) := i;
         if (i /= m) then

--     .......... interchange rows and columns of ar and ai ..........
            for j in mm1 .. n loop
               Swap(ar(i,j),ar(m,j));
               Swap(ai(i,j),ai(m,j));
            end loop;
--
            for j in 1 .. igh loop
               Swap(ar(j,i),ar(j,m));
               Swap(ai(j,i),ai(j,m));
            end loop;
         end if;

--     .......... end interchange ..........
         if (xr /= ScalZero) or (xi /= ScalZero) then
            mp1 := m + 1;
            for i in mp1 .. igh loop
               yr := ar(i,mm1);
               yi := ai(i,mm1);
               if (yr /= ScalZero) or (yi /= ScalZero) then
                  cdiv(yr,yi,xr,xi,yr,yi);
                  ar(i,mm1) := yr;
                  ai(i,mm1) := yi;
                  for j in m .. n loop
                     ar(i,j) := ar(i,j) - yr * ar(m,j) + yi * ai(m,j);
                     ai(i,j) := ai(i,j) - yr * ai(m,j) - yi * ar(m,j);
                  end loop;
                  for j in 1 .. igh loop
                     ar(j,m) := ar(j,m) + yr * ar(j,i) - yi * ai(j,i);
                     ai(j,m) := ai(j,m) + yr * ai(j,i) + yi * ar(j,i);
                  end loop;
               end if;
            end loop;
         end if;
      end loop;
   end comhes;

   procedure comlr2(n,low,igh: in Integer; int: in Pivot; hr,hi: in out Matrix;
                    wr,wi: in out Vector; zr,zi: in out Matrix; ierr: out Integer) is

-- translated from eispack routine comlr2, http://www.netlib.org/eispack/
      ii,l,m,en,im1,itn,its,mp1,enm1: Integer;
      si,sr,ti,tr,xi,xr,yi,yr,zzi,zzr,orm,tst1,tst2: Scalar;

   begin
      ierr := 0;
--     .......... initialize eigenvector matrix ..........
      for i in  1 .. n loop
         for j in 1 .. n loop
            zr(i,j) := ScalZero;
            zi(i,j) := ScalZero;
         end loop;
         zr(i,i) := ScalOne;
      end loop;
--     .......... form the matrix of accumulated transformations
--                from the information left by comhes ..........
      for i in reverse low+1 .. igh-1 loop
         for k in i+1 .. igh loop
            zr(k,i) := hr(k,i-1);
            zi(k,i) := hi(k,i-1);
         end loop;
         ii := int(i);
         if (i /= ii) then
            for k in i .. igh loop
               zr(i,k) := zr(ii,k);
               zi(i,k) := zi(ii,k);
               zr(ii,k) := ScalZero;
               zi(ii,k) := ScalZero;
            end loop;
            zr(ii,i) := ScalOne;
         end if;
      end loop;

--     .......... store roots isolated by cbal ..........
      for i in 1 .. n loop
         if (i < low) or (i > igh) then
            wr(i) := hr(i,i);
            wi(i) := hi(i,i);
         end if;
      end loop;

      en := igh;
      tr := ScalZero;
      ti := ScalZero;
      itn := 30*n;

--     .......... search for next eigenvalue ..........
  <<L220>> if (en < low) then goto L680; end if;

      its := 0;
      enm1 := en - 1;

--     .......... look for single small sub-diagonal element
      loop
         for ll in reverse low .. en loop
            l := ll;
            exit when (l = low);
            tst1 := abs(hr(l-1,l-1)) + abs(hi(l-1,l-1)) + abs(hr(l,l)) + abs(hi(l,l));
            tst2 := tst1 + abs(hr(l,l-1)) + abs(hi(l,l-1));
            exit when (tst2 = tst1);
         end loop;

         if (l = en) then goto L660; end if;

         if (itn = 0) then
--     .......... set error -- all eigenvalues have not
--                converged after 30*n iterations ..........
           ierr := en;
           return;
         end if;

--     .......... form shift ..........
         if (its /= 10) and (its /= 20) then
            sr := hr(en,en);
            si := hi(en,en);
            xr := hr(enm1,en) * hr(en,enm1) - hi(enm1,en) * hi(en,enm1);
            xi := hr(enm1,en) * hi(en,enm1) + hi(enm1,en) * hr(en,enm1);
            if (xr /= ScalZero) or (xi /= ScalZero) then
               yr := (hr(enm1,enm1) - sr) / ScalTwo;
               yi := (hi(enm1,enm1) - si) / ScalTwo;
               csroot(Sqr(Yr)-Sqr(Yi)+xr,ScalTwo*yr*yi+xi,zzr,zzi);
               if (yr * zzr + yi * zzi < ScalZero) then
                  zzr := -zzr;
                  zzi := -zzi;
               end if;

               cdiv(xr,xi,yr+zzr,yi+zzi,xr,xi);
               sr := sr - xr;
               si := si - xi;
            end if;
         else
--     .......... form exceptional shift ..........
            sr := abs(hr(en,enm1)) + abs(hr(enm1,en-2));
            si := abs(hi(en,enm1)) + abs(hi(enm1,en-2));
         end if;

         for i in low .. en loop
            hr(i,i) := hr(i,i) - sr;
            hi(i,i) := hi(i,i) - si;
         end loop;

         tr := tr + sr;
         ti := ti + si;
         its := its + 1;
         itn := itn - 1;

--     .......... look for two consecutive small
--                sub-diagonal elements ..........
         xr := abs(hr(enm1,enm1)) + abs(hi(enm1,enm1));
         yr := abs(hr(en,enm1)) + abs(hi(en,enm1));
         zzr := abs(hr(en,en)) + abs(hi(en,en));

         for mm in reverse l .. enm1 loop
            m := mm;
            exit when (m = l);
            yi := yr;
            yr := abs(hr(m,m-1)) + abs(hi(m,m-1));
            xi := zzr;
            zzr := xr;
            xr := abs(hr(m-1,m-1)) + abs(hi(m-1,m-1));
            tst1 := zzr / yi * (zzr + xr + xi);
            tst2 := tst1 + yr;
            exit when (tst2 = tst1);
         end loop;

--     .......... triangular decomposition h:=l*r ..........
         mp1 := m + 1;

         for i in mp1 .. en loop
            im1 := i - 1;
            xr := hr(im1,im1);
            xi := hi(im1,im1);
            yr := hr(i,im1);
            yi := hi(i,im1);

            if (abs(xr) + abs(xi) < abs(yr) + abs(yi)) then
--     .......... interchange rows of hr and hi ..........
               for j in im1 .. n loop
                  zzr := hr(im1,j);
                  hr(im1,j) := hr(i,j);
                  hr(i,j) := zzr;
                  zzi := hi(im1,j);
                  hi(im1,j) := hi(i,j);
                  hi(i,j) := zzi;
               end loop;

               cdiv(xr,xi,yr,yi,zzr,zzi);
               wr(i) := ScalOne;
            else
               cdiv(yr,yi,xr,xi,zzr,zzi);
               wr(i) := -ScalOne;
            end if;

            hr(i,im1) := zzr;
            hi(i,im1) := zzi;

            for j in i .. n loop
               hr(i,j) := hr(i,j) - zzr * hr(im1,j) + zzi * hi(im1,j);
               hi(i,j) := hi(i,j) - zzr * hi(im1,j) - zzi * hr(im1,j);
            end loop;
         end loop;

--     .......... composition r*l:=h ..........
         for j in mp1 .. en loop
            xr := hr(j,j-1);
            xi := hi(j,j-1);
            hr(j,j-1) := ScalZero;
            hi(j,j-1) := ScalZero;

--     .......... interchange columns of hr, hi, zr, and zi, if necessary ..........
            if (wr(j) > ScalZero) then
               for i in 1 .. j loop
                  Swap(hr(i,j-1),hr(i,j));
                  Swap(hi(i,j-1),hi(i,j));
               end loop;

               for i in low .. igh loop
                  Swap(zr(i,j-1),zr(i,j));
                  Swap(zi(i,j-1),zi(i,j));
               end loop;
            end if;

            for i in 1 .. j loop
               hr(i,j-1) := hr(i,j-1) + xr * hr(i,j) - xi * hi(i,j);
               hi(i,j-1) := hi(i,j-1) + xr * hi(i,j) + xi * hr(i,j);
            end loop;

--     .......... accumulate transformations ..........
            for i in low .. igh loop
               zr(i,j-1) := zr(i,j-1) + xr * zr(i,j) - xi * zi(i,j);
               zi(i,j-1) := zi(i,j-1) + xr * zi(i,j) + xi * zr(i,j);
            end loop;
         end loop;
      end loop;

--     .......... a root found ..........
  <<L660>> null;
      hr(en,en) := hr(en,en) + tr;
      wr(en) := hr(en,en);
      hi(en,en) := hi(en,en) + ti;
      wi(en) := hi(en,en);
      en := enm1;
      goto L220;

--     .......... all roots found.  backsubstitute to find
--                vectors of upper triangular form ..........
  <<L680>> null;
      orm := ScalZero;
      for i in 1 .. n loop
         for j in i .. n loop
            tr := abs(hr(i,j)) + abs(hi(i,j));
            if (tr > orm) then orm := tr; end if;
         end loop;
      end loop;

      hr(1,1) := orm;
      if (n = 1) or IsZero(orm) then return; end if;

      for nn in reverse 2 .. n loop
         en := nn;
         xr := wr(en);
         xi := wi(en);
         hr(en,en) := ScalOne;
         hi(en,en) := ScalZero;
         enm1 := en - 1;

         for i in reverse 1 .. enm1 loop
            zzr := ScalZero;
            zzi := ScalZero;

            for j in i+1 .. en loop
               zzr := zzr + hr(i,j) * hr(j,en) - hi(i,j) * hi(j,en);
               zzi := zzi + hr(i,j) * hi(j,en) + hi(i,j) * hr(j,en);
            end loop;

            yr := xr - wr(i);
            yi := xi - wi(i);
            if IsZero(yr) and then IsZero(yi) then
               tst1 := orm;
               yr := Tst1;
               loop
                  yr := Scal(1.0E-2) * yr;
                  tst2 := orm + yr;
                  exit when (tst2 <= tst1);
               end loop;
            end if;
            cdiv(zzr,zzi,yr,yi,hr(i,en),hi(i,en));

--     .......... overflow control ..........
            tr := abs(hr(i,en)) + abs(hi(i,en));
            if (tr /= ScalZero) then
               tst1 := tr;
               tst2 := tst1 + ScalOne/tst1;
               if (tst2 <= tst1) then
                  for j in i .. en loop
                     hr(j,en) := hr(j,en)/tr;
                     hi(j,en) := hi(j,en)/tr;
                  end loop;
               end if;
            end if;

         end loop;
      end loop;

--     .......... end backsubstitution ..........
--     .......... vectors of isolated roots ..........
      for i in 1 .. n loop
         if (i < low) or (i > igh) then
            for j in i .. n loop
               zr(i,j) := hr(i,j);
               zi(i,j) := hi(i,j);
            end loop;
         end if;
      end loop;

--     .......... multiply by transformation matrix to give
--                vectors of original full matrix.
      for j in reverse low .. n loop
         m := IMin(j,igh);
--
         for i in low .. igh loop
            zzr := ScalZero;
            zzi := ScalZero;
            for k in low .. m loop
               zzr := zzr + zr(i,k) * hr(k,j) - zi(i,k) * hi(k,j);
               zzi := zzi + zr(i,k) * hi(k,j) + zi(i,k) * hr(k,j);
            end loop;
            zr(i,j) := zzr;
            zi(i,j) := zzi;
         end loop;
      end loop;
   end comlr2;

   procedure SortEigen(Eps: in Flt; Wr,Wi: in out Vector; Vr,Vi: in out Matrix) is
     Less1: constant Scalar := ScalOne-Scal(Eps);
     X,Y: Integer;
     Ri,Rj: Scalar;
     A: Pivot(Wr'Range);
     U: Vector(Wr'Range);

     function GT(I,J: Integer) return Boolean is
     begin
       Ri := sqr(Wr(I))+sqr(Wi(I));
       Rj := sqr(Wr(J))+sqr(Wi(J));
       if Less1*Ri > Rj then return  True; end if;
       if Less1*Rj > Ri then return False; end if;
       if Less1*Abs(Wr(I)) > Abs(Wr(J)) then return  True; end if;
       if Less1*Abs(Wr(J)) > Abs(Wr(I)) then return False; end if;
       if Wi(I) > Wi(J) then return  True; end if;
       return False;
     end GT;

     procedure Qsort(L,R: Integer) is
       I,J: integer;
     begin
       I:=L; J:=R; X:=A((L+R)/2);
       loop
         while GT(A(I),X) loop I:=I+1; end loop;
         while GT(X,A(J)) loop J:=J-1; end loop;
         if I <= J then
           Y := A(I); A(I) := A(J); A(J) := Y;
           I:=I+1; J:=J-1;
         end if;
         exit when I>J;
       end loop;
       if L<J then Qsort(L,J); end if;
       if I<R then Qsort(I,R); end if;
     end Qsort;

   begin
     for J in Wr'Range loop A(J) := J; end loop;
     Qsort(Wr'First,Wr'Last);

     U := Wr;
     for J in Wr'Range loop Wr(J) := U(A(J)); end loop;
     U := Wi;
     for J in Wr'Range loop Wi(J) := U(A(J)); end loop;

     for I in Wr'Range loop
       for J in Wr'Range loop U(J) := Vr(I,J); end loop;
       for J in Wr'Range loop Vr(I,J) := U(A(J)); end loop;
       for J in Wr'Range loop U(J) := Vi(I,J); end loop;
       for J in Wr'Range loop Vi(I,J) := U(A(J)); end loop;
     end loop;
   end SortEigen;

   procedure NormalizeEvec(Vr,Vi: in out Matrix) is
     K: Integer;
     S,E,C,Br,Bi: Scalar;
   begin
     for J in Vr'Range(2) loop
       S := ScalZero;
       E := ScalZero;
       for I in Vr'Range(1) loop
         C := Sqr(Vr(I,J))+Sqr(Vi(I,J));
         S := S+C;
         if C>E then E := C; K := I; end if;
       end loop;
       Br := Vr(K,J);
       Bi := Vi(K,J);
       S := Sqrt(S*(Sqr(Br)+Sqr(Bi)));
       Br := Br/S;
       Bi := Bi/S;
       for I in Vr'Range(1) loop
         S := Vr(I,J);
         C := Vi(I,J);
         Vr(I,J) := Br*S+Bi*C;
         Vi(I,J) := Br*C-Bi*S;
       end loop;
       Vi(K,J) := ScalZero;
     end loop;
   end NormalizeEvec;

  procedure Eigen(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix) is
    pragma Suppress(Storage_Check);
    Eps: constant Flt := 1.0E-8;
    N: constant Integer := A'Last(1);
    Low,Igh,Ierr: Integer;
    Int: Pivot(1 .. N);
    Scale: Vector(1 .. N);
    Ar,Ai: Matrix(1 .. N,1 .. N);
  begin
    TraceEnter("Vectors.Num.Eigen");
    Ar := A;
    SetZero(Ai);
    Cbal(N,Ar,Ai,Low,Igh,Scale);
    Comhes(N,Low,Igh,Ar,Ai,Int);
    Comlr2(N,Low,Igh,Int,Ar,Ai,Wr,Wi,Vr,Vi,Ierr);
    if Ierr>0 then
      Show("Error: Problem with eigenvalue number ",Ierr);
    end if;
    Cbabk2(N,Low,Igh,N,Scale,Vr,Vi);
    SortEigen(Eps,Wr,Wi,Vr,Vi);
    NormalizeEvec(Vr,Vi);
    TraceLeave;
  end Eigen;

  procedure FixEval(Wr,Wi: in out Vector; E: in out Scalar) is
--- assuming complex conjugate eigenvalues are next to each other
     J: Integer := Wr'First;
     E1,E2: Scalar;
   begin
     E := ScalZero;
     while J < Wr'Last loop
       E1 := abs(Wi(J));
       E2 := Max(abs(Wr(J)-Wr(J+1)),abs(Wi(J)+Wi(J+1)));
       if E1<E2 then
         Wi(J) := ScalZero;
         E := Max(E,E1/(abs(Wr(J))+E1));
         J := J+1;
       else
         Wr(J) := ScalHalf*(Wr(J)+Wr(J+1));
         Wr(J+1) := Wr(J);
         Wi(J) := ScalHalf*(Wi(J)-Wi(J+1));
         Wi(J+1) := -Wi(J);
         E := Max(E,E2/(abs(Wr(J))+E1));
         J := J+2;
       end if;
     end loop;
     if J=Wr'Last then
       E1 := abs(Wi(J));
       Wi(J) := ScalZero;
       E := Min(E,E1/(abs(Wr(J))+E1));
     end if;
   end FixEval;

   procedure FixEvec(Wi: in Vector; Vr,Vi: in out Matrix; E: in out Scalar) is
     J: Integer;
   begin
     E := ScalZero;
     for I in Wi'Range loop
       J := Wi'First;
       while J <= Wi'Last loop
         if Wi(J)=ScalZero then
           E := Max(E,abs(Vi(I,J)));
           Vi(I,J) := ScalZero;
           J := J+1;
         else
           E := Max(E,Max(abs(Vr(I,J)-Vr(I,J+1)),abs(Vi(I,J)+Vi(I,J+1))));
           Vr(I,J) := ScalHalf*(Vr(I,J)+Vr(I,J+1));
           Vr(I,J+1) := Vr(I,J);
           Vi(I,J) := ScalHalf*(Vi(I,J)-Vi(I,J+1));
           Vi(I,J+1) := -Vi(I,J);
           J := J+2;
         end if;
       end loop;
     end loop;
   end FixEvec;

  procedure Cleanup(W: in out Vector) is
    Small: constant Flt := 1.0E-200;
  begin
    for I in W'Range loop
      if Sup(abs(W(I)))<Small then W(I) := ScalZero; end if;
    end loop;
  end Cleanup;

  procedure Cleanup(V: in out Matrix) is
    Small: constant Flt := 1.0E-100;
  begin
    for I in V'Range(1) loop
      for J in V'Range(2) loop
        if Sup(abs(V(I,J)))<Small then V(I,J) := ScalZero; end if;
      end loop;
    end loop;
  end Cleanup;

  procedure CleanEigen(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix) is
    Err: constant Flt := 1.0E-10;
    E: Scalar;
  begin
    TraceEnter("Vectors.Num.CleanEigen");
    Eigen(A,Wr,Wi,Vr,Vi);
    FixEval(Wr,Wi,E);
    Show("Vectors.Num.CleanEigen warning: could not fix",Sup(E),1);
    FixEvec(Wi,Vr,Vi,E);
    Show("Vectors.Num.CleanEigen warning: could not fix",Sup(E),1);
    Cleanup(Wr); Cleanup(Wi);
    Cleanup(Vr); Cleanup(Vi);
    TraceLeave;
  end CleanEigen;

  procedure ClosestEigen(A: in Matrix; Er,Ei: in out Scalar; Ur,Ui: in out Vector) is
    N: constant Integer := A'Last(1);
    J0: Integer := 0;
    D0: Scalar := Scal(1.0E+99);
    D: Scalar;
    Vr,Vi: Matrix(1 .. N,1 .. N);
  begin
    Eigen(A,Ur,Ui,Vr,Vi);
    for J in 1 .. N loop
      D := Sqr(Ur(J)-Er)+Sqr(Ui(J)-Ei);
      if D<D0 then D0 := D; J0 := J; end if;
    end loop;
    Er := Ur(J0);
    Ei := Ui(J0);
    Ur := GetColumn(J0,Vr);
    Ui := GetColumn(J0,Vi);
  end ClosestEigen;

end Vectors.Num;
