with Ints, Flts;
use Ints, Flts;

generic

  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with function Sup(S: Scalar) return Flt is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;

package Vectors.Num is

  function CountLarge(B: Vector; Cut: Scalar) return Integer;                  -- Number of elements larger than Cut
  procedure Solve(A: in Matrix; X: in out Vector);                             -- solve A*X_out=X_in
  procedure TSolve(A: in Matrix; X: in out Vector);                            -- solve transpose(A)*X_out=X_in
  procedure Invert(A: in out Matrix);                                          -- invert A
  procedure Eigen(A: in Matrix; W: in out Vector);                             -- eigenvalues of symmetric A
  procedure Eigen(A: in Matrix; W: in out Vector; V: in out Matrix);           -- eigenvalues/eigenvectors of symmetric A
  procedure Eigen(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix);   -- eigenvalues/eigenvectors of general A
  procedure CleanEigen(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix);      -- also cleanup
  procedure ClosestEigen(A: in Matrix; Er,Ei: in out Scalar; Ur,Ui: in out Vector);
  procedure Cleanup(W: in out Vector);
  procedure Cleanup(V: in out Matrix);

  Matrix_Singular: exception;
  Getting_Tired: exception;

private

  subtype Pivot is IntVector;

-- procedures translated/adapted from linpack

  procedure dgefa(n: in Integer; a: in out Matrix; ipvt: out Pivot; info: out Integer);
  procedure dgesl(n,job: in Integer; ipvt: in Pivot; a: in Matrix; b: in out Vector);
  procedure dgedi(n,job: in Integer; ipvt: in Pivot; a: in out Matrix; work: in out Vector; det: in out Scalar);

-- procedures translated/adapted from eispack

  function dsign(a,b: Scalar) return Scalar;

  procedure tred1(n: in Integer; a: in out Matrix; d,e,e2: in out Vector);
  procedure tred2(n: in Integer; a: in Matrix; d,e: in out Vector; z: in out Matrix);
  procedure tql1(n: in Integer; d,e: in out Vector; ierr: out Integer);
  procedure tql2(n: in Integer; d,e: in out Vector; z: in out Matrix; ierr: out Integer);

  procedure SortEigen(W: in out Vector; V: in out Matrix);

  procedure cdiv(ar,ai,br,bi: in Scalar; cr,ci: in out Scalar);
  procedure csroot(xr,xi: in Scalar; yr,yi: in out Scalar);
  procedure cbal(n: in Integer; ar,ai: in out Matrix; low,igh: out Integer; scale: in out Vector);
  procedure cbabk2(n,low,igh,m: in Integer; scale: in Vector; zr,zi: in out Matrix);
  procedure comhes(n,low,igh: in Integer; ar,ai: in out Matrix; int: out Pivot);
  procedure comlr2(n,low,igh: in Integer; int: in Pivot; hr,hi: in out Matrix;
                   wr,wi: in out Vector; zr,zi: in out Matrix; ierr: out Integer);

  procedure SortEigen(Eps: in Flt; Wr,Wi: in out Vector; Vr,Vi: in out Matrix);
  procedure NormalizeEvec(Vr,Vi: in out Matrix);
  procedure FixEval(Wr,Wi: in out Vector; E: in out Scalar);
  procedure FixEvec(Wi: in Vector; Vr,Vi: in out Matrix; E: in out Scalar);

end Vectors.Num;
