with Ints, Flts;
use Ints, Flts;

generic

  with function Max(S,T: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Flt is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Intersection(S1,S2: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;

package Vectors.Ops is

  subtype Pivot is IntVector;

  function Norm1(V: Vector) return Scalar;
  function Norm1(A: Matrix) return Scalar;
  procedure Normalize2(V: in out Vector);
  procedure Identity(A: in out Matrix);
  function Identity(N: Positive) return Matrix;
  procedure Symmetrize(A: in out Matrix);
  procedure CheckSymm(A: in out Matrix);
  function AntiSymm(A: in Matrix) return Matrix;
  procedure Invert(D: in out Diag);
  function Inverse(D: in Diag) return Diag;

--- triangular matrices
  procedure LTriangSolve(A: in Matrix; V: in out Vector);             -- A^{-1}*V
  procedure UTriangSolve(A: in Matrix; V: in out Vector);             -- A^{-1}*V
  procedure LTriangInv(A: in Matrix; B: in Matrix; X: in out Matrix); -- solve A*X=B
  procedure UTriangInv(A: in Matrix; B: in Matrix; X: in out Matrix); -- solve A*X=B

  type LUData(N: Positive) is     -- represents A=P*L*U*Q^{-1}
    record
      P,Q: Pivot(1..N);           -- defines permutations
      InvL,U: Matrix(1..N,1..N);  -- lower-triangular and upper-triangular matrices
    end record;

  type DL(N: Positive) is         -- see SimDiag and Cholesky below
    record
      D: Diag(1..N);
      L: Matrix(1..N,1..N);
    end record;

  procedure LDLTransposed(D: in Diag; L: in Matrix ; A: in out Matrix); -- A := L*D*Transpose(L)
  procedure Convert(T: in DL; A: in out Matrix);                        -- A := L*D*Transpose(L)
  procedure Write(Name: in String; T: in DL; Decimal: in Boolean := False);
  procedure Write(Name: in String; T: in LUData; Decimal: in Boolean := False);
  function Read(Name: String; Decimal: Boolean := False) return DL;
  function Read(Name: String; Decimal: Boolean := False) return LUData;

  type SimDiag is new DL;         -- represents A=L*D*L^{-1}
  type Cholesky is new DL;        -- represents A=L*D*Transpose(L), L lower triangular

  procedure Orthogonalize(A: in out Matrix);                         -- Gram-Schmidt
  function OrthoErr(A: in Matrix) return Matrix;                     -- Transpose(A)*A-Id
  procedure Decomp(A: in Matrix; T: in out Cholesky);                -- Cholesky decompose A
  procedure Decomp(A: in Matrix; T: in out LUData);                  -- LU decompose A
  function Decomp(A: Matrix) return LUData;                          -- LU decompose A
  procedure Permute(P: in Pivot; V: in out Vector);                  -- P*V
  procedure UnPermute(P: in Pivot; V: in out Vector);                -- P^{-1}*V
  procedure Mult(T: in LUData; V: in out Vector);                    -- T*V
  procedure Solve(T: in LUData; V: in out Vector);                   -- T^{-1}*V
  procedure Convert(T: in LUData; A: in out Matrix);                 -- T -> A
  procedure Invert(T: in Cholesky; A: in out Matrix);                -- T -> A^{-1}
  procedure Invert(T: in LUData; A: in out Matrix);                  -- T -> A^{-1}
  procedure Solve(A: in Matrix; B: in Vector; X: in out Vector);     -- solve A*X=B
  procedure Invert(A: in out Matrix);                                -- invert A
  procedure InvertSym(A: in out Matrix);                             -- invert symmetric A


private

  procedure FindMax(A: in Matrix; K: in Integer; II,JJ: out Integer);

end Vectors.Ops;
