with Ints, Ints.IO;
use Ints, Ints.IO;

package body Vectors is

  function IsZero(V: Vector) return Boolean is
  begin
    for I in V'Range loop
      if not IsZero(V(I)) then return False; end if;
    end loop;
    return True;
  end IsZero;

  function "="(U,V: Vector) return Boolean is
  begin
    for I in V'Range loop
      if not (U(I)=V(I)) then return False; end if;
    end loop;
    return True;
  end "=";

  function EffLast(V: Vector) return Integer is
  begin
    for I in reverse V'Range loop
      if not IsZero(V(I)) then return I; end if;
    end loop;
    return V'First-1;
  end EffLast;

  procedure SetZero(V: in out Vector) is
  begin
    for I in V'Range loop
      SetZero(V(I));
    end loop;
  end SetZero;

  function ZeroVector(F,L: Integer) return Vector is
    V: Vector(F..L);
  begin
    SetZero(V);
    return V;
  end ZeroVector;

  procedure Neg(V: in out Vector) is
  begin
    for I in V'Range loop
      Neg(V(I));
    end loop;
  end Neg;

  function "-"(V: Vector) return Vector is
    W: Vector(V'Range);
  begin
    for I in V'Range loop
      W(I) := -V(I);
    end loop;
    return W;
  end "-";

  procedure Add(V1: in Vector; V2: in out Vector) is
  begin
    for I in V1'Range loop
      Add(V1(I),V2(I));
    end loop;
  end Add;

  procedure Sum(V1,V2: in Vector; V3: in out Vector) is
  begin
    for I in V1'Range loop
      V3(I) := V1(I)+V2(I);
    end loop;
  end Sum;

  function "+"(V1,V2: Vector) return Vector is
    V3: Vector(V1'Range);
  begin
    Sum(V1,V2,V3);
    return V3;
  end "+";

  procedure Sub(V1: in Vector; V2: in out Vector) is
  begin
    for I in V1'Range loop
      Sub(V1(I),V2(I));
    end loop;
  end Sub;

  procedure Diff(V1,V2: in Vector; V3: in out Vector) is
  begin
    for I in V1'Range loop
      V3(I) := V1(I)-V2(I);
    end loop;
  end Diff;

  function "-"(V1,V2: Vector) return Vector is
    V3: Vector(V1'Range);
  begin
    Diff(V1,V2,V3);
    return V3;
  end "-";

  procedure Mult(S: in Scalar; V: in out Vector) is
  begin
    for I in V'Range loop
      Mult(S,V(I));
    end loop;
  end Mult;

  procedure Prod(S: in Scalar; V1: in Vector; V2: in out Vector) is
  begin
    for I in V1'Range loop
      V2(I) := S*V1(I);
    end loop;
  end Prod;

  function "*"(S: Scalar; V: Vector) return Vector is
    W: Vector(V'Range);
  begin
    Prod(S,V,W);
    return W;
  end "*";

  procedure AddProd(S: in Scalar; V1: in Vector; V2: in out Vector) is
  begin
    if not IsZero(S) then
      for I in V1'Range loop
        AddProd(S,V1(I),V2(I));
      end loop;
    end if;
  end AddProd;

  function IProd(V1,V2: Vector) return Scalar is
    S: Scalar;
  begin
    SetZero(S);
    for I in IMax(V1'First,V2'First) .. IMin(V1'Last,V2'Last) loop
      AddProd(V1(I),V2(I),S);
    end loop;
    return S;
  end IProd;

  procedure Show1(N: in String; V: in Vector; NewLine: in Boolean := True) is
    W: constant Positive := 4;
    VZero: Boolean := True;
  begin
    for I in V'Range loop
      if not IsZero(V(I)) then
        VZero := False;
        Show1(N & Strng(I,W) & "  ",V(I));
      end if;
    end loop;
    if VZero then Show0(N & " Zero_Vector"); end if;
    if NewLine then New_Line; end if;
  end Show1;

  procedure Show2(N: in String; V1,V2: in Vector; NewLine: in Boolean := True) is
    W: constant Positive := 4;
  begin
    for I in V1'Range loop
      if not (IsZero(V1(I)) and IsZero(V2(I))) then
        Show2(N & Strng(I,W) & "  ",V1(I),V2(I));
      end if;
    end loop;
    if NewLine then New_Line; end if;
  end Show2;

  procedure Put(F: in File_Type; V: in Vector; Decimal: in Boolean := False) is
    L: constant Integer := EffLast(V);
    D: constant Integer := L-V'First+1;
  begin
    Put(F,D);
    for I in V'First .. L loop
      Put(F,V(I),Decimal);
   end loop;
  end Put;

  procedure Get(F: in File_Type; V: in out Vector; Decimal: in Boolean := False) is
    Cut: Boolean := Decimal;
    J: Integer := V'First;
    D: Integer;
    S: Scalar;
  begin
    SetZero(V);
    Get(F,D);
    if D>0 then
      for I in 1 .. D loop
        Get(F,S,Decimal);
        if J <= V'Last then V(J) := S; else Cut := True; end if;
        J := J+1;
      end loop;
      if Cut then
        Put_Line("Warning: Input truncated or inaccurate");  -- e.g. decimal
      end if;
    end if;
  end Get;

  function Get(F: File_Type; VFirst: Integer; Decimal: Boolean := False) return Vector is
    L: Integer;
  begin
    Get(F,L);
    declare
      VLast: constant Integer := VFirst+L-1;
      V: Vector(VFirst..VLast);
    begin
      for M in VFirst .. VLast loop
        Get(F,V(M),Decimal);
      end loop;
      return V;
    end;
  end Get;

  procedure Write(Name: in String; V: in Vector; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Writing " & Name);
    Create(F,Out_File,Name);
    Put(F,V,Decimal);
    Close(F);
  end Write;

  procedure Read(Name: in String; V: in out Vector; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_line("Reading " & Name);
    Open(F,In_File,Name);
    Get(F,V,Decimal);
    Close(F);
  end Read;

------------

  procedure SetZeroRow(I: in Integer; A: in out Matrix) is
  begin
    for J in A'Range(2) loop
      SetZero(A(I,J));
    end loop;
  end SetZeroRow;

  procedure SetRow(I: in Integer; V: in Vector; A: in out Matrix) is
  begin
    for J in A'Range(2) loop
      A(I,J) := V(J);
    end loop;
  end SetRow;

  procedure GetRow(I: in Integer; A: in Matrix; V: in out Vector) is
  begin
     for J in A'Range(2) loop
      V(J) := A(I,J);
    end loop;
  end GetRow;

  function GetRow(I: Integer; A: Matrix) return Vector is
     V: Vector(A'Range(2));
  begin
    for J in A'Range(2) loop
      V(J) := A(I,J);
    end loop;
    return V;
  end GetRow;

  procedure SwapRows(I1,I2: in Integer; A: in out Matrix) is
  begin
    if I1 /= I2 then
      for J in A'Range(2) loop
        Swap(A(I1,J),A(I2,J));
      end loop;
    end if;
  end SwapRows;

  procedure SetZeroColumn(J: in Integer; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      SetZero(A(I,J));
    end loop;
  end SetZeroColumn;

  procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      A(I,J) := V(I);
    end loop;
  end SetColumn;

  procedure GetColumn(J: in Integer; A: in Matrix; V: in out Vector) is
  begin
    for I in A'Range(1) loop
      V(I) := A(I,J);
    end loop;
  end GetColumn;

  function GetColumn(J: Integer; A: Matrix) return Vector is
     V: Vector(A'Range(1));
  begin
    for I in A'Range(1) loop
      V(I) := A(I,J);
    end loop;
    return V;
  end GetColumn;

  procedure SwapColumns(J1,J2: in Integer; A: in out Matrix) is
  begin
    if J1 /= J2 then
      for I in A'Range(1) loop
        Swap(A(I,J1),A(I,J2));
      end loop;
    end if;
  end SwapColumns;

  procedure Transpose(A: in out Matrix) is
    AFirst: constant Integer := IMin(A'First(1),A'First(2));
    ALast: constant Integer := IMax(A'Last(1),A'Last(2));
  begin
    for I in AFirst .. ALast loop
      for J in AFirst .. I-1 loop
        Swap(A(I,J),A(J,I));
      end loop;
    end loop;
  end Transpose;

  function Transpose(A: Matrix) return Matrix is
    B: Matrix(A'Range(2),A'Range(1));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        B(J,I) := A(I,J);
      end loop;
    end loop;
    return B;
  end Transpose;

------------

  function IsZero(A: Matrix) return Boolean is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        if not IsZero(A(I,J)) then return False; end if;
      end loop;
    end loop;
    return True;
  end IsZero;

  function EffLast1(A: Matrix) return Integer is
  begin
    for I in reverse A'Range(1) loop
      for J in A'Range(2) loop
         if not IsZero(A(I,J)) then return I; end if;
      end loop;
    end loop;
    return A'First(1)-1;
  end EffLast1;

  function EffLast2(A: Matrix) return Integer is
  begin
    for J in reverse A'Range(2) loop
      for I in A'Range(1) loop
         if not IsZero(A(I,J)) then return J; end if;
      end loop;
    end loop;
    return A'First(2)-1;
  end EffLast2;

  procedure SetZero(A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        SetZero(A(I,J));
      end loop;
    end loop;
  end SetZero;

  function ZeroMatrix(F1,L1,F2,L2: Integer) return Matrix is
    A: Matrix(F1..L1,F2..L2);
  begin
    SetZero(A);
    return A;
  end ZeroMatrix;

  procedure Neg(A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        Neg(A(I,J));
      end loop;
    end loop;
  end Neg;

  function "-"(A: Matrix) return Matrix is
    B: Matrix(A'Range(1),A'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        B(I,J) := -A(I,J);
      end loop;
    end loop;
    return B;
  end "-";

  procedure Add(S: in Scalar; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      Add(S,A(I,I));
    end loop;
  end Add;

  procedure Add(D: in Diag; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      Add(D(I),A(I,I));
    end loop;
  end Add;

  procedure Add(A1: in Matrix; A2: in out Matrix) is
  begin
    for I in A1'Range(1) loop
      for J in A1'Range(2) loop
        Add(A1(I,J),A2(I,J));
      end loop;
    end loop;
  end Add;

  procedure Sum(A1,A2: in Matrix; A3: in out Matrix) is
  begin
    for I in A1'Range(1) loop
      for J in A1'Range(2) loop
        A3(I,J) := A2(I,J)+A1(I,J);
      end loop;
    end loop;
  end Sum;

  function "+"(A1,A2: Matrix) return Matrix is
    A3: Matrix(A1'Range(1),A1'Range(2));
  begin
    Sum(A1,A2,A3);
    return A3;
  end "+";

  procedure Sub(A1: in Matrix; A2: in out Matrix) is
  begin
    for I in A1'Range(1) loop
      for J in A1'Range(2) loop
        Sub(A1(I,J),A2(I,J));
      end loop;
    end loop;
  end Sub;

  procedure Diff(A1,A2: in Matrix; A3: in out Matrix) is
  begin
    for I in A1'Range(1) loop
      for J in A1'Range(2) loop
        A3(I,J) := A2(I,J)-A1(I,J);
      end loop;
    end loop;
  end Diff;

  function "-"(A1,A2: Matrix) return Matrix is
    A3: Matrix(A1'Range(1),A1'Range(2));
  begin
    Diff(A1,A2,A3);
    return A3;
  end "-";

  procedure Mult(S: in Scalar; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        Mult(S,A(I,J));
      end loop;
    end loop;
  end Mult;

  procedure Prod(S: in Scalar; A1: in Matrix; A2: in out Matrix) is
  begin
    for I in A1'Range(1) loop
      for J in A1'Range(2) loop
        A2(I,J) := S*A1(I,J);
      end loop;
    end loop;
  end Prod;

  function "*"(S: Scalar; A: in Matrix) return Matrix is
    B: Matrix(A'Range(1),A'Range(2));
  begin
    Prod(S,A,B);
    return B;
  end "*";

  procedure AddProd(S: in Scalar; A1: in Matrix; A2: in out Matrix) is
  begin
    if not IsZero(S) then
      for I in A1'Range(1) loop
        for J in A1'Range(2) loop
          AddProd(S,A1(I,J),A2(I,J));
        end loop;
      end loop;
    end if;
  end AddProd;

  procedure Mult(D: in Diag; W: in out Vector) is
  begin
    for I in W'Range loop
      Mult(D(I),W(I));
    end loop;
  end Mult;

  procedure Mult(A: in Matrix; W: in out Vector) is
    S: Scalar;
    V: constant Vector := W;
  begin
    for I in V'Range loop
      S := A(I,V'First)*V(V'First);
      for J in V'First+1 .. V'Last loop
        S := S+A(I,J)*V(J);
      end loop;
      W(I) := S;
    end loop;
  end Mult;

  function "*"(A: Matrix; V: Vector) return Vector is
    S: Scalar;
    W: Vector(A'Range(1));
  begin
    for I in W'Range loop
      S := A(I,V'First)*V(V'First);
      for J in V'First+1 .. V'Last loop
        S := S+A(I,J)*V(J);
      end loop;
      W(I) := S;
    end loop;
    return W;
  end "*";

  function "*"(V: Vector; A: Matrix) return Vector is
    S: Scalar;
    W: Vector(A'Range(2));
  begin
    for I in W'Range loop
      S := V(V'First)*A(V'First,I);
      for J in V'First+1 .. V'Last loop
        S := S+V(J)*A(J,I);
      end loop;
      W(I) := S;
    end loop;
    return W;
  end "*";

  procedure Mult(D1: in Diag; D2: in out Diag) is
  begin
    for I in D2'Range loop
      Mult(D1(I),D2(I));
    end loop;
  end Mult;

  procedure Mult(D: in Diag; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        Mult(D(I),A(I,J));
      end loop;
    end loop;
  end Mult;

  procedure Mult(A: in out Matrix; D: in Diag) is
  begin
    for J in A'Range(2) loop
      for I in A'Range(1) loop
        Mult(D(J),A(I,J));
      end loop;
    end loop;
  end Mult;

  procedure Mult(A1: in Matrix; A2: in out Matrix) is
    S: Scalar;
    V: Vector(A2'Range(1));
  begin
    for J in V'Range loop
      for K in V'Range loop
        V(K) := A2(K,J);
      end loop;
      for I in V'Range loop
        S := A1(I,V'First)*V(V'First);
        for K in V'First+1 .. V'Last loop
          S := S+A1(I,K)*V(K);
        end loop;
        A2(I,J) := S;
      end loop;
    end loop;
  end Mult;

  procedure Prod(A1,A2: in Matrix; A3: in out Matrix) is
    KMin: constant Integer := IMin(A1'First(2),A2'First(1));
    KMax: constant Integer := IMax(A1'Last(2),A2'Last(1));
    S: Scalar;
  begin
    for I in A1'Range(1) loop
      for J in A2'Range(2) loop
        S := A1(I,KMin)*A2(KMin,J);
        for K in KMin+1 .. KMax loop
          S := S+A1(I,K)*A2(K,J);
        end loop;
        A3(I,J) := S;
      end loop;
    end loop;
  end Prod;

  function "*"(D: Diag; A: Matrix) return Matrix is
    B: Matrix := A;
  begin
    Mult(D,B);
    return B;
  end "*";

  function "*"(A: Matrix; D: Diag) return Matrix is
    B: Matrix := A;
  begin
    Mult(B,D);
    return B;
  end "*";

  function "*"(A1,A2: Matrix) return Matrix is
    A3: Matrix(A1'Range(1),A2'Range(2));
  begin
    Prod(A1,A2,A3);
    return A3;
  end "*";

  procedure Show1(N: in String; A: in Matrix; NewLine: in Boolean := True) is
    W: constant Positive := 4;
    AZero: Boolean := True;
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        if not IsZero(A(I,J)) then
          AZero := False;
          Show1(N & Strng(I,W) & Strng(J,W) & "  ",A(I,J));
        end if;
      end loop;
    end loop;
    if AZero then Show0(N & " Zero_Matrix"); end if;
    if NewLine then New_Line; end if;
  end Show1;

  procedure Show2(N: in String; A1,A2: in Matrix; NewLine: in Boolean := True) is
    W: constant Positive := 4;
  begin
    for I in A1'Range(1) loop
      for J in A1'Range(2) loop
        if not (IsZero(A1(I,J)) and IsZero(A2(I,J))) then
          Show2(N & Strng(I,W) & Strng(J,W) & "  ",A1(I,J),A2(I,J));
        end if;
      end loop;
    end loop;
    if NewLine then New_Line; end if;
  end Show2;

  procedure Put(F: in File_Type; A: in Matrix; Decimal: in Boolean := False) is
    L: constant Integer := EffLast2(A);
    D: constant Integer := L-A'First(2)+1;
    V: Vector(A'Range(1));
  begin
    Put(F,D);
    for J in A'First(2) .. L loop
      V := GetColumn(J,A);
      Put(F,V,Decimal);
    end loop;
  end Put;

  procedure Get(F: in File_Type; A: in out Matrix; Decimal: in Boolean := False) is
    Cut: Boolean := Decimal;
    J: Integer := A'First(2);
    D: Integer;
    V: Vector(A'Range(1));
  begin
    SetZero(A);
    Get(F,D);
    if D>0 then
      for I in 1 .. D loop
        Get(F,V,Decimal);
        if J <= A'Last(2) then SetColumn(J,V,A); else Cut := True; end if;
        J := J+1;
      end loop;
      if Cut then
        Put_Line("Warning: Input truncated or inaccurate");
      end if;
    end if;
  end Get;

  procedure Write(Name: in String; A: in Matrix; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Writing " & Name);
    Create(F,Out_File,Name);
    Put(F,A,Decimal);
    Close(F);
  end Write;

  procedure Read(Name: in String; A: in out Matrix; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Reading " & Name);
    Open(F,In_File,Name);
    Get(F,A,Decimal);
    Close(F);
  end Read;

end Vectors;
