with Ada.Text_IO;
use Ada.Text_IO;

generic

  type Scalar is private;

  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S,T: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

package Vectors is

  type Vector is array(Integer range <>) of Scalar;

  function IsZero(V: Vector) return Boolean;                           -- V=0
  function "="(U,V: Vector) return Boolean;                           -- V=U
  function EffLast(V: Vector) return Integer;                          -- effective Last
  procedure SetZero(V: in out Vector);                                    -- V := 0
  function ZeroVector(F,L: Integer) return Vector;                     -- return 0
  procedure Neg(V: in out Vector);                                     -- V := -V
  function "-"(V: Vector) return Vector;                               -- return -V
  procedure Add(V1: in Vector; V2: in out Vector);                     -- V2 := V2+V1
  procedure Sum(V1,V2: in Vector; V3: in out Vector);                  -- V3 := V1+V2
  function "+"(V1,V2: Vector) return Vector;                           -- return V1+V2
  procedure Sub(V1: in Vector; V2: in out Vector);                     -- V2 := V2-V1
  procedure Diff(V1,V2: in Vector; V3: in out Vector);                  -- V3 := V1-V2
  function "-"(V1,V2: Vector) return Vector;                           -- return V1-V2
  procedure Mult(S: in Scalar; V: in out Vector);                      -- V := S*V
  procedure Prod(S: in Scalar; V1: in Vector; V2: in out Vector);      -- V2 := S*V1
  function "*"(S: Scalar; V: Vector) return Vector;                    -- return S*V
  procedure AddProd(S: in Scalar; V1: in Vector; V2: in out Vector);   -- V2 := V2+S*V1
  function IProd(V1,V2: Vector) return Scalar;                         -- dot product
  function "*"(V1,V2: Vector) return Scalar renames IProd;

  procedure Show1(N: in String; V: in Vector; NewLine: in Boolean := True);
  procedure Show2(N: in String; V1,V2: in Vector; NewLine: in Boolean := True);
  procedure Put(F: in File_Type; V: in Vector; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; V: in out Vector; Decimal: in Boolean := False);
  function Get(F: File_Type; VFirst: Integer; Decimal: Boolean := False) return Vector;
  procedure Write(Name: in String; V: in Vector; Decimal: in Boolean := False);
  procedure Read(Name: in String; V: in out Vector; Decimal: in Boolean := False);

  type Diag is new Vector;
  type Matrix is array(Integer range <>, Integer range <>) of Scalar;

  procedure SetZeroRow(I: in Integer; A: in out Matrix);
  procedure SetRow(I: in Integer; V: in Vector; A: in out Matrix);
  procedure GetRow(I: in Integer; A: in Matrix; V: in out Vector);
  function GetRow(I: Integer; A: Matrix) return Vector;
  procedure SwapRows(I1,I2: in Integer; A: in out Matrix);
  procedure SetZeroColumn(J: in Integer; A: in out Matrix);
  procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix);
  procedure GetColumn(J: in Integer; A: in Matrix; V: in out Vector);
  function GetColumn(J: Integer; A: Matrix) return Vector;
  procedure SwapColumns(J1,J2: in Integer; A: in out Matrix);
  procedure Transpose(A: in out Matrix);
  function Transpose(A: Matrix) return Matrix;
  function IsZero(A: Matrix) return Boolean;                           -- A=0
  function EffLast1(A: Matrix) return Integer;                         -- effective Last(1)
  function EffLast2(A: Matrix) return Integer;                         -- effective Last(2)
  procedure SetZero(A: in out Matrix);                                 -- A := 0
  function ZeroMatrix(F1,L1,F2,L2: Integer) return Matrix;             -- return 0
  procedure Neg(A: in out Matrix);                                     -- A := -A
  function "-"(A: Matrix) return Matrix;                               -- return -A
  procedure Add(S: in Scalar; A: in out Matrix);                       -- A := S*Id+A
  procedure Add(D: in Diag; A: in out Matrix);                         -- A := D+A
  procedure Add(A1: in Matrix; A2: in out Matrix);                     -- A2 := A2+A1
  procedure Sum(A1,A2: in Matrix; A3: in out Matrix);                  -- A3 := A1+A2
  function "+"(A1,A2: Matrix) return Matrix;                           -- return A1+A2
  procedure Sub(A1: in Matrix; A2: in out Matrix);                     -- A2 := A2-A1
  procedure Diff(A1,A2: in Matrix; A3: in out Matrix);                 -- A3 := A2-A1
  function "-"(A1,A2: Matrix) return Matrix;                           -- return A2-A1
  procedure Mult(S: in Scalar; A: in out Matrix);                      -- A := S*A
  procedure Prod(S: in Scalar; A1: in Matrix; A2: in out Matrix);      -- A2 := S*A1
  function "*"(S: Scalar; A: in Matrix) return Matrix;                 -- return S*A
  procedure AddProd(S: in Scalar; A1: in Matrix; A2: in out Matrix);   -- A2 := A2+S*A1

  procedure Mult(D: in Diag; W: in out Vector);                        -- W := D*W
  procedure Mult(A: in Matrix; W: in out Vector);                      -- W := A*W
  function "*"(A: Matrix; V: Vector) return Vector;                    -- return A*V
  function "*"(V: Vector; A: Matrix) return Vector;                    -- return Transpose(A)*V
  procedure Mult(D1: in Diag; D2: in out Diag);                        -- D2 := D1*D2
  procedure Mult(D: in Diag; A: in out Matrix);                        -- A := D*A
  procedure Mult(A: in out Matrix; D: in Diag);                        -- A := A*D
  procedure Mult(A1: in Matrix; A2: in out Matrix);                    -- A2 := A1*A2
  procedure Prod(A1,A2: in Matrix; A3: in out Matrix);                 -- A3 := A1*A2
  function "*"(D: Diag; A: Matrix) return Matrix;                      -- return D*A
  function "*"(A: Matrix; D: Diag) return Matrix;                      -- return A*D
  function "*"(A1,A2: Matrix) return Matrix;                           -- return A1*A2

  procedure Show1(N: in String; A: in Matrix; NewLine: in Boolean := True);
  procedure Show2(N: in String; A1,A2: in Matrix; NewLine: in Boolean := True);
  procedure Put(F: in File_Type; A: in Matrix; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; A: in out Matrix; Decimal: in Boolean := False);
  procedure Write(Name: in String; A: in Matrix; Decimal: in Boolean := False);
  procedure Read(Name: in String; A: in out Matrix; Decimal: in Boolean := False);

  pragma Inline (Sum,Diff,Prod);

end Vectors;
